import re
import inspect
from collections import defaultdict

from .compiled import Node_Op, Node_Field, Node_Unary, Node_Literal, Node_Function, Node_Op_Is, \
    Node_List, Node_Select, Node_InSelect, Node_Case, Node_Cast
from .exceptions import SqlException_NotImplemented
from .parser import SqlParser, TableRefFinder
from .select_stmt import SqlSelect
from .misc import sql_enquote_id, str_to_var_name
from .eval_utils import SqlEvaluationContext
from .expr_analysis import find_referenced_fields, find_groupby_fields, find_select_fields_pre_post, \
    find_orderby_fields, is_wildcard_all

EXTEND_SQL_RANDOM_SAMPLES = "random_samples"
MODE_QUERY_NUMEXPR = "query.numexpr"
MODE_QUERY_PYTHON = "query.python"
MODE_CALC_PANDAS = "calc.pandas"
MODE_CALC_DASK = "calc.dask"


class ExtendedSqlSelect(SqlSelect):
    def __init__(self):
        super(ExtendedSqlSelect, self).__init__()
        self.random_samples = None


class ExtendedSqlParser(SqlParser):
    """
    We extend SQL grammar a little to take advantage of some pandas functions like describe() and random sampling.

    NOTE: PostGres supports random sampling with its "TABLESAMPLE SYSTEM (n)" expression.
    """
    def __init__(self, sql):
        super(ExtendedSqlParser, self).__init__(sql, allow_any_function=True)
        self.compiled_class = ExtendedSqlSelect
        self.FROM_RESERVED_KEYWORDS.append(EXTEND_SQL_RANDOM_SAMPLES)

    def _terms(self):
        out = super(ExtendedSqlParser, self)._terms()
        for i, term in enumerate(out):
            if term[0] == "order_by":
                out.insert(i, ("random_samples", lambda: self._keyword_then(EXTEND_SQL_RANDOM_SAMPLES, self._parse_random_samples)))
                break
        return out

    def parse_read_statement(self):
        """
        Extension point for parsers that support read statements other than SELECT, like SHOW or DESCRIBE.
        """
        out = self.parse_describe()
        if not out:
            out = self.parse_select_only()
        return out

    def parse_describe(self):
        """
        We commandeer the DESCRIBE verb to access pandas.describe().

        DESCRIBE { dtypes() | percentiles(0.00, 0.05, 0.10, ...) } table_expression
        """
        if not self.skip_keyword("describe"):
            return
        out = SqlDescribe()
        while True:
            if self.skip_keyword("percentiles") and self.peek_token(0) and self.peek_token(0).value == "(":
                # TODO values should be numeric
                out.percentiles = self.parse_expr_or_list()
                if not isinstance(out.percentiles, Node_List):
                    self.report_error("expected list for 'percentiles'")
            elif self.skip_keyword("dtypes") and self.peek_token(0) and self.peek_token(0).value == "(":
                # TODO values are ignored - error if any given
                self.parse_expr_or_list()
                out.dtypes = True
            else:
                break
        # TODO parse include columns options
        out.include = "all"
        # parse table expression
        out.table_expr = self.parse_from_expr()
        return out

    def _parse_random_samples(self):
        """
        Parse custom 'random selection' SQL.  Format is...
            RANDOM_SAMPLES n
              If (n) is a positive integer it reflects the number of samples to take.  If it is a float between 0
              and 1, it is a percentage.
            RANDOM_SAMPLES PCT n
              n/100 is interpreted as above.
            RANDOM_SAMPLES PCT n STATE=123
              the 'state' value can be specified, and leads to a reproducible set of samples
        """
        rand_sel = RandomSamples()
        m = 1
        if self.skip_keyword('pct'):
            m = 0.01
        expr = self.expect_expr()
        rand_sel.count_or_percent = expr.eval(None)
        rand_sel.count_or_percent *= m
        if self.skip_keyword("state"):
            self.skip_punct('=', error_if_missing=True)
            rand_sel.random_state = self.expect_expr().eval(None)
        return rand_sel


class SqlDescribe(object):
    def __init__(self):
        # which tables - a 'FromExpr'
        self.table_expr = None
        self.percentiles = None
        self.include = None
        self.dtypes = False

    def find_all_tables(self):
        """
        Enumerate referenced tables.
        """
        sel = SqlSelect()
        sel.from_expr = self.table_expr
        refs = TableRefFinder()
        refs.refs_in_select(sel)
        return refs.out.keys()


class RandomSamples(object):
    """ Representation of our custom "RANDOM_SAMPLES" SQL extension """
    def __init__(self):
        self.count_or_percent = 0
        self.random_state = 0


class SqlToPandas(object):
    """
    Parses SQL and 'compiles' it into pandas/dask dataframe code.
    """
    def __init__(self, sql: str, local_tables: (callable, str)="tables", ext_table_access: callable=None,
                 tmpvar: str="_",
                 select=None, ext_vars=None, parent=None, external_processing: dict=None, pandas_only: bool=False):
        """
        :param sql:                 SQL to compile.
        :param local_tables:        Name of a {} where local tables are stored, blank for local variables, or a callable to generate code.
        :param ext_table_access:    Code generator for access to external tables.
        :param tmpvar:              The name of the variable used to store the query-in-progress.
        :param select:              The parsed SELECT statement to compile.
        :param ext_vars:            Variables in the SELECT statement reference with "@name".
        :param parent:              Parent instance, indicating this is a subquery.
        :param external_processing: Indicates caller's willingness to do some of the work.  Specifically the
                                    field 'post_limit', if supplied, will be filled in with a number of rows from
                                    a simple LIMIT.
        :param pandas_only:         Flag indicating we can simplify code and assume all dataframes are pandas (not dask).
        """
        self.tmpvar = tmpvar
        self.sql = sql
        self.parent = parent
        self.external_processing = external_processing or {}
        self.pandas_only = pandas_only
        self.code = ""
        self.before = []
        self.after = []
        self.misc_imports = set()
        self.depth = 0 if not parent else parent.depth + 1
        self.referenced_tables = set()
        # when columns are changed, i.e. df[new_col] += 1, we risk changing the underlying table
        self.columns_modified = False
        self.local_tables = local_tables
        self.ext_table_access = ext_table_access
        if isinstance(ext_vars, SqlEvaluationContext):
            self.eval_context = ext_vars
        else:
            self.eval_context = SqlEvaluationContext(ext_vars=ext_vars)
        if select:
            self.compile_select(select)
            return
        parser = ExtendedSqlParser(sql)
        descr = parser.parse_describe()
        if descr:
            self.compile_describe(descr)
            return
        self.compile_select(parser.parse_select_only())
        # final step: make code a bit more readable
        self.more_readable()
        # prepend imports
        if self.misc_imports:
            self.code = "\n".join(f"import {pkg}" for pkg in sorted(self.misc_imports)) + "\n" + self.code

    def compile_describe(self, descr):
        # fill in arguments
        args = []
        if descr.percentiles:
            args.append("percentiles=[%s]" % ", ".join(map(str, descr.percentiles.children)))
        if descr.include:
            args.append("include=%s" % repr(descr.include))
        # start building description
        expr = ""
        if not self.pandas_only:
            expr += inspect.getsource(_to_pandas)
        expr += f"{self.tmpvar} = {self._process_from(descr.table_expr)}\n"
        expr += f"_d = {self.tmpvar}.describe({', '.join(args)})"
        # TODO append() is deprecated - have to detect dask/pandas and use pandas.concat() or dask.dataframe.concat()
        if self.pandas_only:
            expr += f".append({self.tmpvar}.isna().sum().to_frame('nulls').transpose())"
        else:
            expr += f".append(_to_pandas({self.tmpvar}.isna().sum().to_frame('nulls')).transpose())"
        if descr.dtypes:
            if self.pandas_only:
                expr += f".append({self.tmpvar}.dtypes.to_frame('dtype').transpose())"
            else:
                expr += f".append(_to_pandas({self.tmpvar}.dtypes.to_frame('dtype')).transpose())"
            expr += "\n_d = _d.astype(str)"
        expr += "\n_d['_metric_'] = _d.index\n"
        expr += f"{self.tmpvar} = _d"
        self._finish(expr)

    def compile_select(self, select):
        select.simplify(self.eval_context)
        # build all the qualifications around the selected table
        self.process_select(select)
        # get the table
        from_expr = self._process_from(select.from_expr)
        if not self.columns_modified:
            # if we don't change any columns a simple slice will do to prevent changing the original
            tbl_expr = "{v} = {from_expr}[:]".format(v=self.tmpvar, from_expr=from_expr)
        elif self.pandas_only:
            tbl_expr = "{v} = {from_expr}.copy()".format(v=self.tmpvar, from_expr=from_expr)
        else:
            # in pandas there is a member function to copy, in dask we use copy.copy
            tbl_expr = \
                "{v} = {from_expr}\n" \
                "{v} = {v}.copy() if hasattr({v}, 'copy') else copy.copy({v})\n".format(
                v=self.tmpvar, from_expr=from_expr
            )
        self._finish(tbl_expr)

    def _finish(self, tbl_expr):
        # put it together
        if "\n" in tbl_expr:
            # multi-line 'from' has to be separated from qualifiers
            self.code = "%s\n%s = %s" % (tbl_expr, self.tmpvar, self.tmpvar) + self.code
        else:
            # single line 'from' expression can be joined to qualifiers
            self.code = tbl_expr + self.code
        self.code = "\n".join(["\n".join(self.before), self.code, "\n".join(reversed(self.after))])
        self.code = re.sub(r'(^|\n)%s = %s(\n|$)' % (self.tmpvar, self.tmpvar), "\n", self.code)
        self.code = self.code.replace("\n\n", "\n")
        self.code = self.code.strip("\n") + "\n"
        if self.sql:
            self.code = "# SQL: %s\n%s" % (repr(self.sql), self.code)

    def access_table(self, table_ref: tuple):
        # local table - lookup in the specified {}
        if len(table_ref) == 1 or not table_ref[-2]:
            table_name = table_ref[-1]
            if not self.local_tables:
                return "%s" % table_name
            if isinstance(self.local_tables, str):
                return "%s[%s]" % (self.local_tables, repr(table_name))
            return self.local_tables(table_name)
        # external table - call supplied handler
        if not self.ext_table_access:
            self.err("external tables not supported: %s" % repr(table_ref))
        setup, code = self.ext_table_access(table_ref)
        if setup and setup not in self.before:
            self.before.append(setup)
        return code

    def sub_inst(self, tmpvar: str, select=None):
        return SqlToPandas(
            sql="", local_tables=self.local_tables, ext_table_access=self.ext_table_access,
            tmpvar=tmpvar, select=select, parent=self
        )

    def _process_from(self, from_expr):
        if not from_expr:
            self.err("FROM clause is required")
        if from_expr.subquery:
            v_wrap = "_%d" % len(self.before)
            sub = self.sub_inst(tmpvar=v_wrap, select=from_expr.subquery)
            self.before.append(sub.code)
            self.after.append("del %s" % v_wrap)
            code = v_wrap
        else:
            self.referenced_tables.add(from_expr.table)
            code = self.access_table(from_expr.table)
        if not from_expr.join:
            return code
        other = self._process_from(from_expr.join)
        # 'join_fields': table > list[str] -- fields being used from each table
        join_fields = defaultdict(list)
        # find columns to join on, verify it's a simple join
        self.split_join_on(from_expr.join.joinExpr, join_fields)
        # figure out which table named in the join is which
        from_tbl = from_expr.table[-1] if from_expr.table else None
        join_tbl = from_expr.join.table[-1] if from_expr.join.table else None
        f_l = join_fields.get(from_expr.alias or from_tbl, join_fields.get(from_tbl))
        f_r = join_fields.get(from_expr.join.alias or join_tbl, join_fields.get(join_tbl))
        if not f_l or not f_r:
            self.err("could not connect join.on fields to tables: %s" % str(from_expr))
        # TODO if 'multi_key' is set, and result would be dask, raise an exception
        multi_key = len(f_l) != 1 or len(f_r) != 1
        if len(f_l) == 1:
            f_l = f_l[0]
        if len(f_r) == 1:
            f_r = f_r[0]
        how = self._find_join_type(from_expr)
        # left has to be dask if right is dask
        code += \
            "\n" \
            "_r = {other}\n".format(other=other)
        if not self.pandas_only:
            self.misc_imports.add("pandas")
            code += \
                "if isinstance({v}, pandas.DataFrame) and 'dask' in _r.__class__.__name__:\n" \
                "    {v} = daskdataframe.from_pandas({v}, chunksize=10000000)\n".format(
                    v=self.tmpvar, other=other
                )
        # finally, do the join
        code += "{v} = {v}.join(_r.set_index({f_r}, drop=False), on={f_l}, how={how}, rsuffix='_r')".format(
            v=self.tmpvar, other=other, f_r=repr(f_r), f_l=repr(f_l), how=repr(how)
        )
        return code

    def _find_join_type(self, from_expr):
        how = None
        for j_type in ["right", "inner", "outer", "left"]:
            if j_type in from_expr.join.joinType:
                how = j_type
                break
        if not how:
            self.err("unsupported join type: %s" % from_expr.join.joinType)
        return how

    def _process_groupby(self, group_by, aggs, non_agg_cols):
        group_fields = []
        for n_lvl, lvl in enumerate(group_by.levels):
            if not lvl[1]:
                self.err("only ascending groupings are allowed")
            group_fields.append("_agx_%d_%d" % (self.depth, n_lvl))
        # reconstruct
        self.misc_imports.add("pandas")
        lines = [
            "",
            "_g = %s.groupby(%s)" % (self.tmpvar, repr(group_fields)),
            # Big dask dataframes get reduced to a single in-memory dataframe
            "%s = pandas.DataFrame()" % self.tmpvar
        ]
        # TODO support large post-aggregation results
        '''
        In order to support large aggregate results...
        1) create new dask dataframe from the new groupby (_g)'s index, specifying number of partitions:
           _ = _g[{group_fields}].first(split_to=..., split_every=...).to_frame()
        2) assign each column as below, but make sure split_to matches in every case
        3) and, of course, we have to actually choose a value for split_out
           could be max(_.npartitions//4, 1) <-- or something like that

        See: https://examples.dask.org/dataframes/02-groupby.html#Many-groups
        '''
        # pull in the aggregated fields
        for agg_var, agg_expr in aggs:
            lhs = "%s[%s]" % (self.tmpvar, repr(agg_var))
            lines.append(self.render_calc(lhs, agg_expr, mode="aggregate"))
        # carry forward all the non-aggregate fields
        for col in non_agg_cols:
            lines.append("%s[%s] = _g[%s].first()" % (self.tmpvar, repr(col), repr(col)))
        lines.append("del _g")
        lines.append("_ = _")
        return "\n".join(lines)

    def _process_groupby_all(self, aggs, non_agg_cols):
        build = []
        if not self.pandas_only and "def _compute(df):" not in self.before:
            self.before += [
                "def _compute(df):",
                "    if hasattr(df, 'compute'):",
                "        return df.compute()",
                "    return df"
            ]
        # pull in the aggregated fields
        for agg_var, agg_expr in aggs:
            node_code = self.render_node(agg_expr, mode="aggregate")
            if not self.pandas_only:
                node_code = f"_compute({node_code})"
            build.append("%s: [%s]" % (repr(agg_var), node_code))
        # carry forward all the non-aggregate fields
        for col in non_agg_cols:
            build.append("%s: _g[%s].head(1)" % (repr(col), repr(col)))
        self.misc_imports.add("pandas")
        lines = [
            "",
            "_g = %s" % self.tmpvar,
            "%s = pandas.DataFrame({%s})" % (self.tmpvar, ", ".join(build)),
            "del _g",
            "_ = _"
        ]
        return "\n".join(lines)

    def _process_orderby(self, order_by, field_map):
        if not order_by:
            return ""
        exprs, orders = zip(*order_by.levels)
        expr_fields = []
        for expr in exprs:
            if not isinstance(expr, Node_Field):
                self.err("only single field supported for order by")
            expr_fields.append((field_map or {}).get(tuple(expr.fieldName), expr.fieldName[-1]))
        fields_repr = repr(expr_fields)
        asc = True
        if all(orders):
            """ all ascending """
            pass
        elif not any(orders):
            """ all descending """
            asc = False
        else:
            self.err("mixed asc/desc not supported yet")
        if self.pandas_only:
            out = ".sort_values({fields}, ascending={asc}).set_index({field1}, drop=False)\n".format(
                    fields=fields_repr, field1=repr(exprs[0].fieldName[-1]), asc=asc
                )
        else:
            self.misc_imports.add("pandas")
            out = \
                "\n" \
                "if isinstance({v}, pandas.DataFrame):\n" \
                "    {v} = {v}.sort_values({fields}, ascending={asc})\n" \
                "else:\n".format(
                    v=self.tmpvar, fields=fields_repr, asc=asc
                )
            if not asc:
                out += "    raise Exception('descending sort not supported in dask mode')\n"
            elif len(exprs) > 1:
                out += "    raise Exception('multi-key sort not supported in dask mode')\n"
            else:
                out += "    {v} = {v}.set_index({field1}, drop=False)\n".format(
                    v=self.tmpvar, field1=repr(exprs[0].fieldName[-1])
                )
            out += "{v} = {v}".format(v=self.tmpvar)
        return out

    def _process_range(self, select):
        start = select.start.eval(self.eval_context) if select.start else 0
        limit = select.limit.eval(self.eval_context) if select.limit else -1
        lo = start or ""
        if start and limit != -1:
            hi = start + limit
        else:
            hi = limit if limit != -1 else ""
        if not lo and hi and "post_limit" in self.external_processing:
            # caller is willing to truncate results
            self.external_processing["post_limit"] = hi
            return ""
        if lo or hi:
            if self.pandas_only:
                out = ".iloc[{lo}:{hi}]".format(lo=lo, hi=hi)
            else:
                self.misc_imports.add("pandas")
                out = \
                    "\n" \
                    "if isinstance({v}, pandas.DataFrame):\n" \
                    "    {v} = {v}.iloc[{lo}:{hi}]\n" \
                    "else:\n".format(
                        v=self.tmpvar, lo=lo, hi=hi
                    )
                if hi:
                    out += \
                        "    {v} = {v}.head({hi}, npartitions=-1, compute=False)\n".format(
                            v=self.tmpvar, hi=hi
                        )
                if lo:
                    # TODO tail() only works on the last partition, so really large chunks will be truncated
                    #  - it would be better to raise an exception than to silently return bad data
                    out += \
                        "    _nr = {v}.shape[0].compute()\n" \
                        "    {v} = {v}.tail(max(_nr - {start}, 0), compute=False)\n".format(
                            v=self.tmpvar, start=start
                        )
                out += "{v} = {v}".format(v=self.tmpvar)
            return out
        return ""

    def _process_random_samples(self, random_samples):
        if not random_samples:
            return ""
        args = []
        n = random_samples.count_or_percent
        if n < 1:
            args.append("frac=%s" % repr(n))
        else:
            args.append("n=%s" % repr(int(n)))
        if random_samples.random_state:
            args.append("random_state=%s" % repr(int(random_samples.random_state)))
        return ".sample(%s)" % ", ".join(args)

    def _process_union(self, selects):
        if not selects:
            return ""
        # TODO protect from nested unions
        prefix = "_u"
        table_vars = []
        code_out = "\n"
        for n_sel, other_sel in enumerate(selects):
            u_var = prefix + "%d" % (n_sel+1)
            tbl_code = SqlToPandas(
                sql="", local_tables=self.local_tables, ext_table_access=self.ext_table_access,
                tmpvar=u_var, select=other_sel, ext_vars=self.eval_context
            ).code
            code_out += tbl_code
            #if not code_out.endswith("\n"):
            #    code_out += "\n"
            table_vars.append(u_var)
        table_vars.append(self.tmpvar)
        # result has to be dask if any sub-tables are dask
        self.misc_imports.add("pandas")
        if self.pandas_only:
            code_out += "{v} = pandas.concat([{tbls}], sort=False)".format(v=self.tmpvar, tbls=', '.join(table_vars))
        else:
            self.misc_imports.add("dask.dataframe")
            code_out += \
                "_union = dask.dataframe.concat if not all(isinstance(sub, pandas.DataFrame) for sub in %s) \\\n" \
                "    else (lambda tbls: pandas.concat(tbls, sort=False))\n" % repr(table_vars)
            code_out += "{v} = _union([{tbls}])\n{v} = {v}".format(
                v=self.tmpvar, tbls=", ".join(table_vars)
            )
        return code_out

    def _process_distinct(self, distinct):
        if distinct is None:
            return ""
        subset = ""
        if len(distinct):
            subset = ", ".join(repr(f[-1]) for f in distinct)
        return ".drop_duplicates(%s)" % subset

    def process_select(self, select):
        # analyze fields used in each part of the statement
        field_analysis = FieldAnalysis(select, depth=self.depth)
        # we might need to carry forward some fields for the WHERE/GROUP/ORDER clauses
        aggs = []
        non_agg_cols = []
        fb = FieldsBuilder(self, aggs, non_agg_cols)
        del_after_where = set()
        del_cols = set()
        field_map = {}
        if not field_analysis.sel_wildcard:
            field_analysis.analyze()
            del_cols = field_analysis.del_cols
            del_after_where = field_analysis.del_after_where
            fb.after_field_map = field_map = field_analysis.field_map
            # calculate group-by expressions
            gb_exprs = field_analysis.analyze_groupby()
            # do calculations, etc.
            self.code += fb.build(select.fields, field_analysis.temp_exprs, gb_exprs)
        if select.where:
            self.code += self.render_query(select.where, field_map=field_map)
        if del_after_where:
            self.code += ".drop(columns=%s)" % repr(del_after_where)
        if select.group_by:
            self.code += self._process_groupby(select.group_by, aggs, non_agg_cols)
        elif aggs:
            self.code += self._process_groupby_all(aggs, non_agg_cols)
        if select.having:
            self.code += self.render_query(select.having, field_map=field_map)
        self.code += self._process_distinct(select.distinct)
        self.code += self._process_random_samples(select.random_samples)
        self.code += self._process_orderby(select.order_by, field_map)
        self.code += self._process_range(select)
        if del_cols:
            # drop the temporary columns used by group/order
            self.code += ".drop(columns=%s)" % repr(del_cols)
        self.code += self._process_union(select.union)

    def err(self, message):
        raise SqlException_NotImplemented(message)

    def _render_op(self, node, mode, field_map):
        op_map_calc = {
            "||": "|", "=": "==", "<>": "!=",
            "and": "&", "or": "|", "&&": "&",
        }
        op_map_query = {
            "||": "|", "=": "==", "<>": "!=",
            "&&": "&",
        }
        op_map = op_map_query if mode.startswith("query") else op_map_calc
        op_bad = {
            "like", "not like", "regexp", "not regexp", "match", "not match"
        }
        op = node.operator
        if op in op_bad:
            self.err("operator not supported yet: %s" % node.operator)
        if op == "in" and mode != MODE_QUERY_NUMEXPR:
            return "%s.isin(%s)" % (
                self.render_node(node.children[0], mode, field_map),
                self.render_node(node.children[1], mode, field_map)
            )
        # checks for NULL
        if op in {Node_Op_Is.OP_IS, Node_Op_Is.OP_IS_NOT} and mode.startswith("query"):
            if isinstance(node.children[1], Node_Literal) and node.children[1].value is None:
                if mode == MODE_QUERY_NUMEXPR:
                    raise CantUseNumExpr()
                return "%s(%s).isnull()" % ("not " if op == Node_Op_Is.OP_IS_NOT else "", self.render_node(node.children[0], mode, field_map))
            # TODO may not make sense - what are other valid uses of 'is'?
            # try mapping to =/!=
            op = {Node_Op_Is.OP_IS: "=", Node_Op_Is.OP_IS_NOT: "!="}[op]
        op = op_map.get(op, op)
        return "(%s)" % (" " + op + " ").join(map(lambda n: self.render_node(n, mode, field_map), node.children))

    def _render_fn(self, node, mode, field_map):
        fn = node.functionName
        if node.isAggregate() and mode == "aggregate":
            return self._render_fn__aggregate(node, fn)
        if mode in {"query", MODE_QUERY_NUMEXPR, MODE_QUERY_PYTHON}:
            return self._render_fn__query(fn, node, mode, field_map)
        elif mode == MODE_CALC_DASK and fn == "if":
            # compensate for dask bug - https://github.com/dask/dask/issues/5839
            self.misc_imports.add("dask.array")
            return "dask.array.where(({arg0}).values, {args})".format(
                arg0=self.render_node(node.children[0], mode, field_map),
                args=", ".join(self.render_node(child, mode, field_map) for child in node.children[1:]),
            )
        elif mode in {MODE_CALC_PANDAS, MODE_CALC_DASK} and fn in NUMPY_FUNCTIONS:
            lib = "numpy" if mode == MODE_CALC_PANDAS else "dask.array"
            self.misc_imports.add(lib)
            return "{lib}.{fn}({args})".format(
                lib=lib,
                fn=NUMPY_FUNCTIONS[fn],
                args=", ".join(self.render_node(child, mode, field_map) for child in node.children),
            )
        elif mode == "calc":
            """
            For 'calc' mode, we start with the set of functions we can use without recourse to numpy/dask.array.
            """
            r = self._render_fn__calc(fn, node, mode, field_map)
            if r:
                return r
        self.err("function not supported yet: %s" % node.functionName)

    def _render_fn__aggregate(self, node, fn):
        """
        Compile aggregation function calls.
        """
        fn = AGGREGATES.get(fn, fn)
        if fn == "count" and len(node.children) == 1 and str(node.children[0]) == "*":
            return "_g.count().min(axis=int(_g.__class__.__name__ != 'DataFrame'))"
        varname = str_to_var_name("_agg_" + str(node.children[0]))
        return "_g[%s].%s()" % (repr(varname), fn)

    def _render_fn__query(self, fn, node, mode, field_map):
        """
        Render functions in the context of pandas.query().
        """
        #  https://numexpr.readthedocs.io/projects/NumExpr3/en/latest/user_guide.html#supported-functions
        if fn in QUERY_FUNCTIONS:
            return "%s(%s)" % (QUERY_FUNCTIONS[fn], ", ".join(self.render_node(child, mode) for child in node.children))
        if fn in QUERY_SUB_FUNCTIONS:
            return "%s.%s(%s)" % (
                self.render_node(node.children[0], mode, field_map),
                QUERY_SUB_FUNCTIONS[fn],
                ", ".join(self.render_node(child, mode, field_map) for child in node.children[1:])
            )
        if fn in (set(NATIVE_FUNCTIONS) | set(NATIVE_PROPERTIES) | set(NUMPY_FUNCTIONS)):
            raise NeedToUseCalc

    def _render_fn__calc(self, fn, node, mode, field_map):
        """
        Render a function call in "calc" mode, which uses overloaded pandas operators to do a piecewise
        query.
        """
        if fn in NATIVE_FUNCTIONS:
            # these functions can be coded identically in pandas or dask
            fn = NATIVE_FUNCTIONS[fn]
            return "(%s).%s(%s)" % (
                self.render_node(node.children[0], mode, field_map),
                fn,
                ", ".join(self.render_node(child, mode, field_map) for child in node.children[1:])
            )
        elif fn in NATIVE_PROPERTIES:
            fn = NATIVE_PROPERTIES[fn]
            return "(%s).%s" % (
                self.render_node(node.children[0], mode, field_map),
                fn
            )
        elif fn in NUMPY_FUNCTIONS:
            # but not these
            raise NeedToBifurcate()

    def _render_case(self, node, mode, field_map: dict=None):
        """
        CASE statement.
        """
        # convert the case statement to a bunch of if() functions
        when_then_else = node.children
        pairs = []
        for n in range(0, len(when_then_else)-1, 2):
            pairs.append((when_then_else[n], when_then_else[n+1]))
        pairs.reverse()
        out = when_then_else[-1] if len(when_then_else) % 2 else Node_Literal(None)
        for cond, if_1 in pairs:
            out = Node_Function("if", [cond, if_1, out])
        # render it
        return self._render_fn(out, mode, field_map)

    def render_query(self, node, field_map: dict=None):
        """
        Apply a filter (i.e. a WHERE clause) to the current table.
        """
        try:
            try:
                return ".query(%s)" % repr(self.render_node(node, mode=MODE_QUERY_NUMEXPR, field_map=field_map))
            except CantUseNumExpr:
                return ".query(%s, engine='python')" % repr(self.render_node(node, mode=MODE_QUERY_PYTHON, field_map=field_map))
        except NeedToUseCalc:
            calc = self.render_calc("_filt", node, field_map=field_map)
            self.after.append("del _filt")
            return "\n" + calc + "{v} = {v}[_filt]".format(v=self.tmpvar)

    def render_calc(self, lhs, node, mode: str="calc", field_map: dict=None):
        """
        Perform a calculation and assign it to the given LHS (left hand side).
        """
        try:
            return "%s = %s\n" % (lhs, self.render_node(node, mode=mode, field_map=field_map))
        except NeedToBifurcate:
            pandas_calc = self.render_node(node, mode=mode+".pandas", field_map=field_map)
            if self.pandas_only:
                return \
                    "{lhs} = {pandas_calc}\n".format(
                    lhs=lhs, pandas_calc=pandas_calc
                )
            self.misc_imports.add("pandas")
            return \
                "if isinstance({v}, pandas.DataFrame):\n" \
                "    {lhs} = {pandas_calc}\n" \
                "else:\n" \
                "    {lhs} = {dask_calc}\n".format(
                    v=self.tmpvar, lhs=lhs,
                    pandas_calc=pandas_calc,
                    dask_calc=self.render_node(node, mode=mode+".dask", field_map=field_map)
                )

    def render_node(self, node, mode: str=MODE_QUERY_NUMEXPR, field_map: dict=None):
        """
        Render an SQL expression.

        x < 4:
          mode==query:    x < 4           (i.e. for use inside DataFrame.query())
          mode==calc:  _['x'] + 1         (i.e. for computing a column)
        """
        if isinstance(node, (Node_Op, Node_Op_Is)):
            return self._render_op(node, mode, field_map)
        if isinstance(node, Node_Field):
            field_name = (field_map or {}).get(tuple(node.fieldName), node.fieldName[-1])
            if mode.startswith("query"):
                return sql_enquote_id(field_name, "mysql")
            else:
                return "%s[%s]" % (self.tmpvar, repr(field_name))
        if isinstance(node, Node_Unary):
            return "%s %s" % (node.operator, self.render_node(node.children[0], mode, field_map=field_map))
        if isinstance(node, Node_Literal):
            return repr(node.value)
        if isinstance(node, Node_Cast):
            pandas_type = SQL_TYPES_TO_PANDAS_TYPES.get(node.to_type, node.to_type)
            return "%s.astype(%s, errors='ignore')" % (
                self.render_node(node.children[0], mode, field_map=field_map), repr(pandas_type)
            )
        if isinstance(node, Node_Function):
            return self._render_fn(node, mode, field_map)
        if isinstance(node, Node_Case):
            return self._render_case(node, mode, field_map)
        if isinstance(node, Node_InSelect):
            v_wrap = "_%d" % len(self.before)
            sub = self.sub_inst(tmpvar=v_wrap, select=node.select_spec)
            self.before.append(sub.code)
            self.before.append("{sub} = set({sub}[{sub}.columns[0]])".format(sub=v_wrap))
            self.after.append("del %s" % v_wrap)
            return self.render_node(node.children[0], mode, field_map=field_map) + ".isin(%s)" % v_wrap
        if isinstance(node, Node_Select):
            v_wrap = "_%d" % len(self.before)
            sub = self.sub_inst(tmpvar=v_wrap, select=node.select_spec)
            self.before.append(sub.code)
            return "{tbl}[{tbl}.columns[0]].max()".format(tbl=v_wrap)
        ''' TODO it isn't clear to me when this is used in a SELECT
        if isinstance(node, Node_List):
            subs = [self.render_node(sub, mode=mode, field_map=field_map) for sub in node.children]
            return "(" + "".join(f"{sub}, " for sub in subs) + ")"
        '''
        self.err("node type not supported yet: %s" % node.__class__.__name__)

    def split_join_on(self, node, out):
        if isinstance(node, Node_Field):
            if len(node.fieldName) < 2:
                self.err("joins require aliases for sub-selects")
            tbl = node.fieldName[-2]
            fld = node.fieldName[-1]
            out[tbl].append(fld)
            return
        elif isinstance(node, Node_Op):
            if node.operator not in {"=", "==", "and"}:
                self.err("unsupported operation for join.on: %s" % node.operator)
        else:
            self.err("unsupported for join.on: %s" % node.__class__.__name__)
        if hasattr(node, "children"):
            for sub in node.children:
                self.split_join_on(sub, out)

    def more_readable(self):
        # copy-all-columns, followed by copy-some-columns, i.e. [:][['a', 'b']]
        self.code = re.sub(r"\[:\](\[\['[^'\n]+'(, '[^'\n]+')*\]\])", r'\1', self.code)


class FieldAnalysis(object):
    def __init__(self, select, depth: int):
        self.select = select
        self.depth = depth
        self.sel_wildcard = is_wildcard_all(select.fields)
        self.del_after_where = set()
        self.del_cols = set()
        self.field_map = {}
        self.temp_exprs = []

    def analyze(self):
        """
        Work out which fields need to be 'carried forward' from our input table, to be used by WHERE, GROUP BY, etc..

        For example...
          SELECT a, b FROM table WHERE c=1

          Here, the 'c' column is not preserved in the field list but it is neededby the WHERE clause.
        """
        sel_in_fields, sel_out_fields = find_select_fields_pre_post(self.select.fields)
        where_fields = find_referenced_fields(self.select.where)
        groupby_fields = find_groupby_fields(self.select.group_by)
        orderby_fields = find_orderby_fields(self.select.order_by)
        preserve_w = where_fields - set(sel_out_fields)
        preserve_g_o = (groupby_fields | orderby_fields) - set(sel_out_fields)
        preserve_o = orderby_fields - set(sel_out_fields)
        preserve = preserve_w | preserve_g_o
        for nf, f in enumerate(preserve):
            fn = "_tmp_%d" % nf
            self.field_map[f] = fn
            if f in preserve_w and f not in preserve_g_o:
                self.del_after_where.add(fn)
            elif f in preserve_o:
                self.del_cols.add(fn)
            self.temp_exprs.append((Node_Field(f), fn))

    def analyze_groupby(self):
        """
        Calculate all the group-by values and store them in temporary columns.
        """
        gb_exprs = []
        for n_level, level in enumerate(self.select.group_by.levels if self.select.group_by else []):
            fn = "_agx_%d_%d" % (self.depth, n_level)
            gb_exprs.append((level[0], fn))
        return gb_exprs


def aggregate_exprs(node, out):
    if isinstance(node, Node_Function) and node.isAggregate():
        expr = node.children[0]
        out[str(expr)] = expr
        return True
    found = False
    if hasattr(node, "children"):
        for sub in node.children:
            found = found or aggregate_exprs(sub, out)
    return found


AGGREGATES = {
    "min": "min",
    "max": "max",
    "count": "count",
    "count_distinct": "nunique",
    "sum": "sum",
    "total": "sum",
    "avg": "mean",
    "average": "mean",
    "median": "median",
    "stdev": "std",
    # TODO more pandas functions: mode(), var()
}


class FieldsBuilder(object):
    """
    Compiles the field selection/calculation of an SQL SELECT, i.e. "select _____ from ..."
    """
    def __init__(self, s2p, aggs, non_agg_cols):
        """
        :param s2p:             The sql-to-pandas compilation that is in progress.
        :param aggs:            All of the fields which are calculated based on aggregation (i.e. the
                                fields whose expressions contain aggregation functions like 'avg' or 'sum')
                                are listed in the provided [].  A [] of (alias, expression) is stored here.
        :param non_agg_cols:    The names of all output fields (i.e. aliases) which are not aggregation
                                expressions are listed here, in the provided [].
        """
        self.s2p = s2p
        self.aggs = aggs
        self.non_agg_cols = non_agg_cols
        self.keep = []
        self.rename = {}
        self.calculations = []
        self.calculations_after = []
        self.agg_exprs = {}
        self.exclude_fields = set()
        self.after_field_map = {}

    def one_field(self, expr, alias, hidden: bool=False, after: bool=False):
        """
        Process one field.
        :param expr:    Expression for field
        :param alias:   Alternate name
        :param hidden:  This is a temporary field which will be removed later.
        """
        if isinstance(expr, Node_Field) and not hidden:
            return self._one_field_simple(expr, alias)
        calc_name = alias or str(expr)
        if aggregate_exprs(expr, self.agg_exprs):
            self.aggs.append((calc_name, expr))
        else:
            lhs = "%s[%s]" % (self.s2p.tmpvar, repr(calc_name))
            if after:
                self.calculations_after.append(self.s2p.render_calc(lhs, expr, field_map=self.after_field_map))
            else:
                self.calculations.append(self.s2p.render_calc(lhs, expr))
            self.s2p.columns_modified = True
            if calc_name not in self.keep and not after:
                self.keep.append(calc_name)
                if not hidden:
                    self.non_agg_cols.append(calc_name)

    def _one_field_simple(self, expr: Node_Field, alias):
        """
        Process the simplest case, where there is no calculation, only the selection of a named
        field, i.e. "SELECT field FROM table".
        :param expr:        The expression, i.e. which field is being selected.
        :param alias:       The alternate name selected for the field, if any.
        """
        field_name = expr.fieldName[-1]
        if field_name.startswith("*["):
            # "SELECT *[removed1, removed2]" is an extension to SQL that lets us exclude some fields without
            # having to know what fields are available.
            # TODO add routine for this in analysis or SqlUtil
            self.exclude_fields = set(map(lambda f: f.strip(), field_name[2:].strip("]").split(",")))
            return
        # indicate this is not a temporary field
        self.keep.append(field_name or alias)
        # indicate this is a non-aggregation field
        self.non_agg_cols.append(alias or field_name)
        # if renaming is taking place, make a note of that
        if alias and field_name != alias:
            self.rename[field_name] = alias

    def build(self, field_list, temp_exprs, temp_exprs_after):
        for expr, alias in field_list:
            self.one_field(expr, alias)
        for expr, alias in (temp_exprs or []):
            self.one_field(expr, alias, hidden=True)
        for expr, alias in (temp_exprs_after or []):
            self.one_field(expr, alias, hidden=True, after=True)
        # save off arguments for all aggregation function calls
        for agg_var, agg_expr in self.agg_exprs.items():
            if agg_var == "*":
                self.keep.append("*")
                continue
            agg_col = str_to_var_name("_agg_" + agg_var)
            lhs = "%s[%s]" % (self.s2p.tmpvar, repr(agg_col))
            self.calculations.append(self.s2p.render_calc(lhs, agg_expr))
            self.keep.append(agg_col)
            self.s2p.columns_modified = True
        code = ""
        if self.calculations:
            code += "\n" + "".join(self.calculations) + "%s = %s" % (self.s2p.tmpvar, self.s2p.tmpvar)
        if self.exclude_fields:
            code += "[list(sorted((set(%s.columns) - %s) | set(%s)))]" % (
                self.s2p.tmpvar, repr(self.exclude_fields), repr(self.keep))
        elif "*" not in self.keep:
            code += "[%s]" % repr(self.keep)
        if self.rename:
            code += ".rename(columns=%s)" % repr(self.rename)
        if self.calculations_after:
            code += "\n" + "".join(self.calculations_after) + "%s = %s" % (self.s2p.tmpvar, self.s2p.tmpvar)
        return code


class NeedToBifurcate(Exception):
    """
    This exception is raised when code needs to be different between pandas and dask.
    """


class NeedToUseCalc(Exception):
    """
    This indicates we can't use pandas.query() and have to use 'calc'-like expressions.  That is...
    """


class CantUseNumExpr(Exception):
    """
    This indicates we can't use query.numexpr(), and have to use the 'python' engine instead.
    """


def _to_pandas(df):
    import pandas
    if isinstance(df, pandas.DataFrame):
        return df
    return df.compute()


# These functions can be run like so:
#    dataframe['column'].abs()
# This is a mapping from the standard SQL function to the qualified function name within pandas/dask.
NATIVE_FUNCTIONS = {
    # string
    "concat": "str.cat",
    "length": "str.len", "len": "str.len", "char_length": "str.len",
    "lcase": "str.lower", "lower": "str.lower", "ucase": "str.upper", "upper": "str.upper",
    "trim": "str.strip", "ltrim": "str.lstrip", "rtrim": "str.rstrip",
    "replace": "str.replace", "substr": "str.slice", "substring": "str.slice",
    # numeric
    "abs": "abs",
    # datetime
    "strftime": "dt.strftime"
}


# these are functions which can be accessed using properties of Series
NATIVE_PROPERTIES = {
    "year": "dt.year", "month": "dt.month", "day": "dt.day"
}


# These functions come from either numpy or dask.array.
NUMPY_FUNCTIONS = {
    # numeric
    'acos': 'arccos', 'asin': 'arcsin', 'atan': 'arctan', 'atan2': 'arctan2', 'ceil': 'ceil', 'ceiling': 'ceil',
    'cos': 'cos', 'degrees': 'degrees', 'exp': 'exp', 'floor': 'floor', 'ln': 'log', 'log': 'log', 'log10': 'log10',
    'log2': 'log2', 'mod': 'mod', 'pi': 'pi', 'pow': 'power', 'power': 'power', 'radians': 'radians', 'rand': 'random',
    'round': 'round', 'sign': 'sign', 'sin': 'sin', 'sqrt': 'sqrt', 'tan': 'tan', 'truncate': 'truncate',
    # other
    'if': 'where'
}


QUERY_FUNCTIONS = {
    "sin": "sin", "cos": "cos", "tan": "tan",
    "asin": "arcsin", "acos": "arccos", "atan": "arctan", "atan2": "arctan2",
    "log": "log", "log10": "log10", "exp": "exp",
    "sqrt": "sqrt", "abs": "abs",
}


QUERY_SUB_FUNCTIONS = NATIVE_FUNCTIONS


SQL_TYPES_TO_PANDAS_TYPES = {"datetime": "datetime64"}



from .parser import SqlParser


def parse_insert(sql: str):
    """
    Parsing of 'INSERT INTO'.  Returns:
      table_name
      field_list
      a generator for the list of rows
    """
    parser = SqlParser(sql)
    if not parser.skip_keyword("INSERT"):
        parser.report_error("expected 'insert'")
    if not parser.skip_keyword("INTO"):
        parser.report_error("expected 'into' after 'insert'")
    table_name = parser.parse_table_name()
    parser.skip_punct("(", error_if_missing=True)
    fields = []
    while True:
        fields.append(parser.any_keyword("field name"))
        if not parser.skip_punct(','):
            break
    if not parser.skip_punct(')'):
        parser.report_error("expected ')' after field list")
    if not parser.skip_keyword("VALUES"):
        parser.report_error("expected values for insert")
    def rows():
        while True:
            parser.skip_punct("(", error_if_missing=True)
            values = []
            while True:
                values.append(parser.parse_expr().eval(None))
                if not parser.skip_punct(','):
                    break
            if not parser.skip_punct(')'):
                parser.report_error("expected ')' after value list")
            yield values
            if not parser.skip_punct(','):
                break
        if parser.cur_token():
            parser.report_error("expected end of statement")
    return table_name, fields, rows()


def parse_create_db(sql: str):
    """
    Check for a CREATE DATABASE statement.  Returns None, or a tuple with information about the statement.
    """
    parser = SqlParser(sql)
    if not parser.skip_keyword("CREATE"):
        return
    if not parser.skip_keyword("DATABASE"):
        return
    if_not_exists = False
    if parser.skip_keyword("IF"):
        if not parser.skip_keyword("NOT"):
            parser.report_error("expected 'not'")
        if not parser.skip_keyword("EXISTS"):
            parser.report_error("expected 'exists'")
        if_not_exists = True
    db_name = parser.any_keyword("database name")
    if parser.cur_token():
        parser.report_error("expected end of statement")
    return db_name, if_not_exists


def parse_delete(sql: str):
    """
    Parsing of 'DELETE'.  Returns:
      table_name
      where clause
    """
    parser = SqlParser(sql)
    if not parser.skip_keyword("DELETE"):
        parser.report_error("expected 'delete'")
    if not parser.skip_keyword("FROM"):
        parser.report_error("expected 'from' after 'delete'")
    table_name = parser.parse_table_name()
    where = None
    if parser.skip_keyword("WHERE"):
        where = parser.expect_expr()
    if parser.cur_token():
        parser.report_error("expected end of statement")
    # TODO limit could also be supported here
    return table_name, where


def parse_drop_table(sql: str):
    """
    Parsing of 'DROP TABLE'.  Returns:
      table_name
      if_exists flag
    """
    parser = SqlParser(sql)
    if not parser.skip_keyword("DROP"):
        parser.report_error("expected 'drop'")
    if not parser.skip_keyword("TABLE"):
        parser.report_error("expected 'table'")
    if_exists = parser.skip_keyword(["IF", "EXISTS"])
    table_name = parser.parse_table_name()
    if parser.cur_token():
        parser.report_error("expected end of statement")
    return table_name, if_exists



import functools

from .compiled import _cmp, Node_Field, Node_Op, SqlOrderBy
from .data import SqlTableStream, SqlFieldInfo
from .eval_utils import find_field_in_list, SqlEvaluationContext


class Union(SqlTableStream):
    def __init__(self, streams):
        SqlTableStream.__init__(self, streams)
        self.fields = None
        self.fieldMap = {}
        self.streams = streams

    def _calcFields(self):
        # align the fields
        new_fields = []
        for stream in self.streams:
            for field in stream.getFields():
                key = (field.alias or field.field).lower()
                if key not in self.fieldMap:
                    self.fieldMap[key] = len(new_fields)
                    new_fields.append(SqlFieldInfo(field=field.field, alias=field.alias))
        self.fields = new_fields

    def getFields(self):
        if self.fields is None:
            self._calcFields()
        return self.fields

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        self.getFields()
        for stream in self.streams:
            for row in stream.getRowIterator():
                new_row = [None, ] * len(self.fields)
                for n, field in enumerate(stream.getFields()):
                    key = (field.alias or field.field).lower()
                    idx = self.fieldMap[key]
                    new_row[idx] = row[n]
                if self.context.debugLevel:
                    print("union: out={0}".format(new_row))
                yield tuple(new_row)

    def canLimit(self, start, count):
        if start or count >= 0:
            return False
        return True

    def canWhere(self, where):
        if where:
            return False
        return True

    def canSort(self, sort):
        if sort:
            return False
        return True


class _FieldCalc(SqlTableStream):
    def __init__(
            self, stream, compiler, field_mappings, alias_to_table=None, table_to_alias=None,
            aggregation_grouper=None, downstream_visible_from_upstream=False, upstream_sort=None
    ):
        SqlTableStream.__init__(self, stream)
        self.oldFields = None
        self.fields = None
        self.stream = stream
        self.compiler = compiler
        self.fieldMappings = field_mappings
        self.aliasToTable = alias_to_table
        self.tableToAlias = table_to_alias
        self.aggregationGrouper = aggregation_grouper
        self.downstreamVisibleFromUpstream = downstream_visible_from_upstream
        self.upstream_sort = upstream_sort
        self.stream_cache = {}

    def _gen_stream_field_ref(self, field, alias):
        """
        Generate a reference to a field in 'stream'.
        """
        table_alias = None
        if self.tableToAlias:
            table_alias = self.tableToAlias.get((field.table,)) \
                          or self.tableToAlias.get((field.database, field.table))
        return SqlFieldInfo(
            table=field.table, field=field.field,
            database=field.database, alias=alias or field.alias,
            table_alias=table_alias
        )

    def _calcFields(self):
        self.oldFields = self.stream.getFields()
        # generate new field list
        new_fields = []
        for n, fm in enumerate(self.fieldMappings):
            expr = fm[0]
            alias = fm[1]
            if isinstance(expr, Node_Field):
                i_old = find_field_in_list(expr.fieldName, self.stream.context)
                if i_old != -1:
                    # found field in 'stream' by name
                    new_field = self._gen_stream_field_ref(self.oldFields[i_old], alias)
                elif alias:
                    # field not found - we'll just use the alias
                    new_field = SqlFieldInfo(table=None, field=None, alias=alias)
                elif isinstance(expr.fieldName, tuple) and len(expr.fieldName) >= 2:
                    # field not found - we'll use the given field name as an alias
                    new_field = SqlFieldInfo(table=expr.fieldName[0], field=None,
                                                    alias=expr.fieldName[-1])
                else:
                    # list the alias, without any real table/field information
                    new_field = SqlFieldInfo(table=None, field=None, alias=expr.fieldName[-1])
            else:
                # calculated field
                new_field = SqlFieldInfo(table=None, field=str(expr), alias=alias)
            new_fields.append(new_field)
        return new_fields

    def supplyContext(self, context):
        if self.aliasToTable:
            context = context.derive(tableAliases=self.aliasToTable)
        SqlTableStream.supplyContext(self, context)

    def getFields(self):
        if self.fields is None:
            self.fields = self._calcFields()
        return self.fields

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        self.getFields()
        if self.aggregationGrouper and not self.stream.canSort(self.upstream_sort):
            i = _FieldCalcRowIter(self, where)
            for row in i.scan_aggregate_upstream_unsorted():
                yield row
        else:
            i = _FieldCalcRowIter(self, where, sort=self.upstream_sort)
            for row in i.scan(sort=sort, start=start, count=count):
                yield row

    def _clone_fieldMappings(self):
        from copy import deepcopy
        return deepcopy(self.fieldMappings)

    def _eval_row(self, upstream_row, aggregate=True, fieldMappings=None):
        fieldMappings = fieldMappings or self.fieldMappings
        row_to_build = [None] * len(self.fieldMappings)
        ctx1 = self.context.derive(fields=self.getFields(), record=row_to_build, aggregate=aggregate)
        ctx2 = ctx1.derive(fields=self.oldFields, record=upstream_row, cache=self.stream_cache)
        for n, fm in enumerate(fieldMappings):
            row_to_build[n] = fm[0].eval(ctx2)
        return tuple(row_to_build)

    def canSort(self, sort):
        if self.aggregationGrouper:
            if not sort or sort == self.upstream_sort:
                return True
            return False
        return super(_FieldCalc, self).canSort(sort)

    def canWhere(self, where):
        return True

    def canLimit(self, start, count):
        if self.aggregationGrouper:
            return False
        return super(_FieldCalc, self).canLimit(start, count)


class _FieldCalcRowIter(object):
    def __init__(self, calc, where=None, sort=None):
        self.calc = calc
        calc.getFields()
        self.ctx_downstream = calc.context.derive(fields=calc.getFields())
        if calc.downstreamVisibleFromUpstream:
            self.ctx_upstream = self.ctx_downstream.derive(fields=calc.oldFields)
        else:
            self.ctx_upstream = SqlEvaluationContext(fields=calc.oldFields)
        where_parts = split_expr_by_upstream_availability(
            where, self.ctx_upstream, self.ctx_downstream, prefer_upstream=bool(calc.aggregationGrouper)
        )
        self.where_keep = where_parts[0]
        self.where_upstream = where_parts[1]
        self.apply_sort = sort
        self.use_stream = calc.stream
        if not calc.stream.canWhere(self.where_upstream):
            self.use_stream = calc.compiler.support_where_and_limit(calc.stream)

    def check_where(self, row):
        if not self.where_keep:
            return True
        ok = test_sql_filter(self.where_keep, self.ctx_downstream.derive(record=row))
        if self.ctx_downstream.debugLevel:
            print("calculateOrSelectFields: where={0}, {1}".format(ok, self.where_keep))
        return ok

    def _dbg(self, **kwargs):
        if self.calc.context and self.calc.context.debugLevel:
            print("calculateOrSelectFields: %s" % kwargs)

    def _tail(self, last_group, last_row):
        if self.calc.aggregationGrouper:
            new_row = self.calc._eval_row(last_row, aggregate=False)
            self._dbg(agg=last_group, row_out=new_row)
            if self.check_where(new_row):
                return new_row

    def scan(self, sort=None, start=0, count=-1):
        calc = self.calc
        prev_group = None
        prev_row = None
        for row in self.use_stream.getRowIterator(where=self.where_upstream, sort=self.apply_sort or sort, start=start, count=count):
            ctx = self.ctx_upstream.derive(record=row)
            if not calc.aggregationGrouper:
                # non-aggregated row
                new_row = calc._eval_row(row)
                self._dbg(row_in=row, row_out=new_row)
                if self.check_where(new_row):
                    yield new_row
                continue
            # determine grouping
            current_group = calc.aggregationGrouper(ctx)
            if prev_group and current_group != prev_group:
                # group changed
                new_row = calc._eval_row(prev_row, aggregate=False)
                self._dbg(agg=current_group, row_out=new_row)
                if self.check_where(new_row):
                    yield new_row
            prev_group = current_group
            # aggregate to current group
            self._dbg(agg=current_group, row_in=row)
            calc._eval_row(row)
            prev_row = row
        out = self._tail(prev_group, prev_row)
        if out:
            yield out

    def scan_aggregate_upstream_unsorted(self):
        """
        Aggregation that doesn't require upstream sorting.
        """
        calc = self.calc
        groups = {}
        for row in self.use_stream.getRowIterator(where=self.where_upstream):
            ctx = self.ctx_upstream.derive(record=row)
            current_group = calc.aggregationGrouper(ctx)
            group = groups.get(current_group)
            if group:
                group_fm = group[0]
            else:
                group_fm = calc._clone_fieldMappings()
                groups[current_group] = (group_fm, row)
            calc._eval_row(row, fieldMappings=group_fm)
        for _, (group_fm, group_row) in sorted(groups.items()):
            row = calc._eval_row(group_row, fieldMappings=group_fm, aggregate=False)
            if self.check_where(row):
                yield row


class _JoinMerger(SqlTableStream):
    def __init__(self, stream_l, stream_r, join_type, join_expr, compiler):
        super(_JoinMerger, self).__init__([stream_l, stream_r])
        self.streamL = stream_l
        self.streamR = stream_r
        self.joinType = join_type
        self.joinExpr = join_expr
        self.compiler = compiler
        self.fields = None
        self.leftJoin = "left" in join_type
        self.rightJoin = "right" in join_type
        self.innerJoin = "inner" in join_type
        self.fullJoin = "full" in join_type or "outer" in join_type
        if not self.leftJoin and not self.rightJoin and not self.innerJoin and not self.fullJoin:
            self.innerJoin = True

    def supplyContext(self, context):
        SqlTableStream.supplyContext(self, context)
        if context:
            self.context = context.derive(fields=self.getFields())

    def getFields(self):
        if self.fields is None:
            self.fields = self.streamL.getFields() + self.streamR.getFields()
        return self.fields

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        expr_parts = analyze_expr_for_join(where_and_where(where, self.joinExpr), self.streamL.context, self.streamR.context)
        s_l, s_r = self._prepare_streams(expr_parts)
        i = _JoinMergerIterator(expr_parts, self, s_l, s_r)
        for row in i.iterate():
            if i.post_filter(row):
                yield row

    def _prepare_streams(self, expr_parts):
        s_l = self.streamL
        s_r = self.streamR
        compiler = self.compiler
        assert s_l.context and s_r.context
        #print("analyze_expr_for_join: {0}".format(where_and_where(where, self.joinExpr)))
        #print("leftExprs={0}, rightExprs={1}, leftOnly={2}, rightOnly={3}".format([str(x) for x in expr_parts.leftExprs], [str(x) for x in expr_parts.rightExprs], expr_parts.leftOnly, expr_parts.rightOnly))
        # pre-filter the parts of the expression that relate to the source streams
        if expr_parts.left_only:
            s_l = compiler.apply_where(s_l, expr_parts.left_only)
        if expr_parts.right_only:
            s_r = compiler.apply_where(s_r, expr_parts.right_only)
        assert s_l.context and s_r.context
        # apply sorting to incoming streams
        if expr_parts.left_exprs:
            order = SqlOrderBy()
            for var in expr_parts.left_exprs:
                order.addLevel(var)
            s_l = compiler.apply_sort(s_l, order)
            order = SqlOrderBy()
            for var in expr_parts.right_exprs:
                order.addLevel(var)
            s_r = compiler.apply_sort(s_r, order)
        return s_l, s_r

    def canLimit(self, start, count):
        if start or count >= 0:
            return False
        return True

    def canWhere(self, where):
        return True

    def canSort(self, sort):
        return not sort


class _JoinMergerIterator(object):
    def __init__(self, expr_parts, joiner, stream_l, stream_r):
        self.expr_parts = expr_parts
        self.joiner = joiner
        self.context = joiner.context
        self.stream_l = stream_l
        self.stream_r = stream_r
        self.fields_l = stream_l.getFields()
        self.fields_r = stream_r.getFields()
        self.null_l = (None,) * len(self.fields_l)
        self.null_r = (None,) * len(self.fields_r)
        self.i_l = iter(self.stream_l.getRowIterator())
        self.i_r = iter(self.stream_r.getRowIterator())
        self.r_l = _nxt(self.i_l)
        self.r_r = _nxt(self.i_r)

    def post_filter(self, r):
        if self.expr_parts.other:
            ok = test_sql_filter(self.expr_parts.other, self.context.derive(record=r))
            if self.context.debugLevel:
                print("join: postFilter={0}, pass={1}".format(self.expr_parts.other, ok))
            if not ok:
                return False
        if self.context.debugLevel:
            print("join: emit={0}".format(r))
        return True

    def compare(self, r_l, r_r):
        c = 0
        for n, x_l in enumerate(self.expr_parts.left_exprs):
            x_r = self.expr_parts.right_exprs[n]
            v_l = x_l.eval(self.context.derive(record=r_l, fields=self.fields_l)) if r_l else None
            v_r = x_r.eval(self.context.derive(record=r_r, fields=self.fields_r)) if r_r else None
            c = _cmp(v_l, v_r)
            if c:
                break
        return c

    def _iter_dbg(self, match):
        if self.context.debugLevel:
            print("join {0}, L={1}, R={2}, match={3}".format(self.joiner.joinType, self.r_l, self.r_r, match))

    def iterate(self):
        while self.r_l and self.r_r:
            match = self.compare(self.r_l, self.r_r)
            self._iter_dbg(match)
            for out in self._cycle(match):
                yield out
        # one of the streams has ended
        for out in self._tail():
            yield out

    def _cycle(self, match):
        # let the recessive stream catch up
        if self.joiner.leftJoin:
            if match > 0:
                # skip right
                self.r_r = _nxt(self.i_r)
                return
            elif match < 0:
                # emit left only
                out = self.r_l + self.null_r
                yield out
                self.r_l = _nxt(self.i_l)
                return
        if self.joiner.rightJoin:
            if match < 0:
                self.r_l = _nxt(self.i_l)
                return
            elif match > 0:
                # emit right only
                out = self.null_l + self.r_r
                yield out
                self.r_r = _nxt(self.i_r)
                return
        if self.joiner.innerJoin:
            if match < 0:
                self.r_l = _nxt(self.i_l)
                return
            elif match > 0:
                self.r_r = _nxt(self.i_r)
                return
        # we must be in full join mode if match!=0 at this point
        if match < 0:
            out = self.r_l + self.null_r
            yield out
            self.r_l = _nxt(self.i_l)
            return
        elif match > 0:
            out = self.null_l + self.r_r
            yield out
            self.r_r = _nxt(self.i_r)
            return
        # merge
        # - gather all matching instances on left and right
        for out in self._gather_leftovers(match):
            yield out

    def _tail(self):
        # one of the streams has ended
        if self.joiner.innerJoin:
            return
        if not self.joiner.rightJoin:
            while self.r_l:
                out = self.r_l + self.null_r
                yield out
                self.r_l = _nxt(self.i_l)
        if not self.joiner.leftJoin:
            while self.r_r:
                out = self.null_l + self.r_r
                yield out
                self.r_r = _nxt(self.i_r)

    def _gather_leftovers(self, match):
        # cross-join for multiple matching rows
        all_l = [self.r_l]
        all_r = [self.r_r]
        while match == 0:
            self.r_l = _nxt(self.i_l)
            if not self.r_l:
                break
            match = self.compare(self.r_l, self.r_r)
            if match == 0:
                all_l.append(self.r_l)
        match = 0
        while match == 0:
            self.r_r = _nxt(self.i_r)
            if not self.r_r:
                break
            match = self.compare(all_l[0], self.r_r)
            if match == 0:
                all_r.append(self.r_r)
        # yield all combinations of all_l and all_r
        for o_l in all_l:
            for o_r in all_r:
                out = o_l + o_r
                yield out


class _UpstreamAvailabilitySplitter(object):
    DOWN = 0
    UP = 1

    def __init__(self, upstream_context, downstream_context, prefer_upstream=False):
        self.upstream_context = upstream_context
        self.downstream_context = downstream_context
        self.prefer_upstream = prefer_upstream
        self.split_out = [None, None]

    def send_node(self, cat, node):
        if self.split_out[cat] is None:
            self.split_out[cat] = node
        else:
            self.split_out[cat] = Node_Op("and", [self.split_out[cat], node])

    def analyze_group(self, node):
        if self.keep_downstream(node):
            self.send_node(self.DOWN, node)
        else:
            self.send_node(self.UP, node)

    def can_go_up(self, node):
        if isinstance(node, Node_Field):
            return self.upstream_context.namedValueAvailable(node.fieldName)
        if hasattr(node, "children") and not all(map(self.can_go_up, node.children)):
            return False
        return True

    def can_go_down(self, node):
        if isinstance(node, Node_Field):
            return self.downstream_context.namedValueAvailable(node.fieldName)
        if hasattr(node, "children") and not all(map(self.can_go_down, node.children)):
            return False
        return True

    def keep_downstream(self, node):
        if self.can_go_up(node):
            return False
        if self.can_go_down(node):
            return True
        if self.prefer_upstream:
            return False
        return True

    def split(self, expr):
        if expr:
            for sub in top_level_ands(expr):
                self.analyze_group(sub)
        return self.split_out


def split_expr_by_upstream_availability(expr, upstream_context, downstream_context, prefer_upstream=False):
    """
    Split an expression into two sets of components to be evaluated by different
    levels of a compiled stream.  Any components that only reference fields in the
    upstreamContext will be put into the second category, and all others go into the first.

    :return:  A list with two elements: the downstream and the local.
    """
    return _UpstreamAvailabilitySplitter(upstream_context, downstream_context, prefer_upstream).split(expr)


def test_sql_filter(where, ctx):
    """
    Evaluate a WHERE expression against a given row
    """
    if not where:
        return True
    v = where.eval(ctx)
    return bool(v)


def where_and_where(where1, where2):
    """
    Combine two where expressions with an AND.
    """
    if not where1 and not where2:
        return None
    if where1 and not where2:
        return where1
    if where2 and not where1:
        return where2
    return Node_Op("and", [where1, where2])


class _JoinAnalyzer(object):
    def __init__(self, left_context, right_context):
        self.out = _JoinParts()
        self.left_context = left_context
        self.right_context = right_context

    def _categorize_field(self, field_name):
        if find_field_in_list(field_name, self.left_context) != -1:
            return "L"
        if find_field_in_list(field_name, self.right_context) != -1:
            return "R"
        return "LR"

    def categorize(self, node):
        if isinstance(node, Node_Field):
            return self._categorize_field(node.fieldName)
        if hasattr(node, "children"):
            t = None
            for child in node.children:
                sub = self.categorize(child)
                if sub:
                    t = sub if (t == sub or t is None) else "LR"
            return t

    def analyze_group(self, node):
        {
            "L": self.out.addLeft,
            "R": self.out.addRight,
            "LR": self.analyze_lr,
            None: self.out.addOther
        }[self.categorize(node)](node)

    def analyze_lr(self, node):
        if not isinstance(node, Node_Op) or node.operator not in {"=", "=="}:
            self.out.addOther(node)
            return
        c_l = self.categorize(node.children[0])
        c_r = self.categorize(node.children[1])
        if c_l == "L" and c_r == "R":
            self.out.left_exprs.append(node.children[0])
            self.out.right_exprs.append(node.children[1])
        elif c_l == "R" and c_r == "L":
            self.out.right_exprs.append(node.children[0])
            self.out.left_exprs.append(node.children[1])
        else:
            self.out.addOther(node)

    def analyze(self, expr):
        for sub in top_level_ands(expr):
            self.analyze_group(sub)
        return self.out


def analyze_expr_for_join(expr, left_context, right_context):
    # TODO <R.field is NONE> does not make sense as a pre-filter because missing fields come through as None
    return _JoinAnalyzer(left_context, right_context).analyze(expr)


class _JoinParts(object):
    """
    leftExprs, rightExprs: aligned lists of expressions that can be used to qualify a join
    leftOnly, rightOnly: expressions that apply only to the left or right table
    other: the part of the expression that can only be used to post-qualify the join
    """

    def __init__(self):
        self.left_exprs = []
        self.right_exprs = []
        self.left_only = None
        self.right_only = None
        self.other = None

    def addLeft(self, node):
        if self.left_only:
            self.left_only = Node_Op("and", [self.left_only, node])
        else:
            self.left_only = node

    def addRight(self, node):
        if self.right_only:
            self.right_only = Node_Op("and", [self.right_only, node])
        else:
            self.right_only = node

    def addOther(self, node):
        if self.other:
            self.other = Node_Op("and", [self.other, node])
        else:
            self.other = node


def _nxt(i):
    try:
        return i.__next__()
    except StopIteration:
        return None


def top_level_ands(node):
    """
    Generate an iteration over all of the ANDs in an expression that precede all other parts, i.e.:
        a AND b  ==>  (a, b)
        a AND (b or c) ==>  (a, (b or c))
        a * 2 ==> ()
    """
    if isinstance(node, Node_Op) and node.operator == "and":
        for sub in node.children:
            for s2 in top_level_ands(sub):
                yield s2
    else:
        yield node


class SortableStream(SqlTableStream):
    """
    Add sorting capability to a stream.
    """
    def __init__(self, stream):
        SqlTableStream.__init__(self, stream)

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        if self.upstream.canSort(sort):
            # let upstream do its own sorting if it is able
            out = self.upstream.getRowIterator(where, sort, start, count)
            # if upstream can sort but can't limit, we have to apply the limit here as well
            if not self.upstream.canLimit(start, count):
                # TODO optimize this for large streams
                out = list(out)
                if count:
                    out = out[start:start+count]
                elif start:
                    out = out[start:]
            return out
        else:
            ctx = self.upstream.context.derive(fields=self.getFields())
            if count >= 0:
                # there is a limit we can apply
                upstream_rows = self.upstream.getRowIterator(where)
                rows = top_sort(upstream_rows, sort=sort, limit=start + count, context=ctx)
                if start:
                    rows = rows[start:]
            else:
                # we have to sort the entire data set
                upstream_rows = self.upstream.getRowIterator(where, None, start, count)
                rows = list(upstream_rows)
                def sort_cmp(a, b):
                    return sort.compare(a, b, ctx)
                rows.sort(key=functools.cmp_to_key(sort_cmp))
                if count >= 0:
                    rows = rows[start:start+count]
                elif start:
                    rows = rows[start:]
            return rows

    def canSort(self, sort):
        return True

    def canLimit(self, start, count):
        return True


class TopList(object):
    def __init__(self, cmp, limit):
        self.cmp = cmp
        self.limit = limit
        self.data = []

    def locate(self, value):
        if not self.data:
            return 0
        lo = 0
        hi = len(self.data)
        while lo < hi:
            mid = (lo + hi) // 2
            cmp = self.cmp(value, self.data[mid])
            if cmp == 0:
                return mid
            if cmp > 0:
                lo = mid + 1
            else:
                hi = mid
        return lo

    def add(self, value):
        ins_before = self.locate(value)
        self.data.insert(ins_before, value)
        if len(self.data) > self.limit:
            self.data = self.data[:self.limit]


def top_sort(upstream, sort, limit, context):
    def sort_cmp(a, b):
        return sort.compare(a, b, context)
    top = TopList(cmp=sort_cmp, limit=limit)
    for row in upstream:
        top.add(row)
    return top.data


class WhereAndLimit(SqlTableStream):
    """
    Adds filtering and range capabilities.
    """
    def __init__(self, stream, force_post_filter):
        SqlTableStream.__init__(self, stream)
        self.force_post_filter = force_post_filter

    def getRowIterator(self, where=None, sort=None, start=0, count=-1):
        limit_upstream = (0, -1)
        where_upstream = None
        if self.upstream.canLimit(start, count):
            limit_upstream = (start, count)
            start, count = 0, -1
        if not self.force_post_filter and self.upstream.canWhere(where):
            where_upstream = where
            where = None
        base_ctx = self.context.derive(fields=self.getFields())
        for row in self.upstream.getRowIterator(where_upstream, sort, limit_upstream[0], limit_upstream[1]):
            ok = test_sql_filter(where, base_ctx.derive(record=row))
            if self.context.debugLevel:
                print("where: {0}, {1}".format(ok, where))
            if not ok:
                continue
            if start:
                start -= 1
                continue
            if count == 0:
                break
            if count > 0:
                count -= 1
            yield row

    def canWhere(self, where):
        return True

    def canLimit(self, start, count):
        return True



"""
Random access reads against dataframes using only reset and forward iterators.
"""
import numpy
import math


class DFIterator(object):
    def __init__(self, df):
        self.df = df
        self.columns = list(df.columns)
        self.re_iter = lambda: df.itertuples(index=False, name=None)
        self.iter = None
        self.pos = 0
        self.eof = False
        self.is_dask = "dask" in str(df.__class__)
        self._n_rows = df.shape[0]

    @property
    def n_rows(self):
        if hasattr(self._n_rows, "compute"):
            self._n_rows = self._n_rows.compute()
        return self._n_rows

    @staticmethod
    def clean(row):
        out = []
        for v in row:
            if hasattr(v, "compute"):
                v = v.compute()
            if v != v:
                v = None
            elif isinstance(v, (numpy.int8, numpy.int16, numpy.int32, numpy.int64)):
                v = int(v)
            elif isinstance(v, (numpy.float16, numpy.float32, numpy.float64)):
                v = float(v)
                if not math.isfinite(v):
                    v = None
            elif isinstance(v, float) and not math.isfinite(v):
                v = None
            out.append(v)
        return tuple(out)

    def fetch(self, start: int=0, end: int=None):
        """
        Fetch a range of results.  Returns a tuple of:
          (0) - column names
          (1) - An [] of []s, values for each row.
        """
        # for pandas we can use 'iloc' to get our slice
        if not self.is_dask:
            # TODO optimize -- this commented out version may actually be faster, but don't take my word for it
            # rows = json.loads(self.df.to_json(orient="values"))
            rows = list(map(self.clean, self.df.iloc[start:end].itertuples(index=False, name=None)))
            return self.columns, rows
        # special case: entire dataset
        if start == 0 and end is None:
            rows = list(map(self.clean, self.df.itertuples(index=False, name=None)))
            return self.columns, rows
        # support negative range with head() and tail()
        if start < 0:
            return self._fetch_from_end(start, end)
        # one row at a time, optimized for paged forward access
        if self.iter is None:
            self.iter = self.re_iter()
            self.pos = 0
        clean = self.clean
        if start < self.pos:
            self.iter = self.re_iter()
            self.pos = 0
            self.eof = False
        rows = []
        if self.eof:
            return self.columns, rows
        try:
            while self.pos < start:
                next(self.iter)
                self.pos += 1
            while end is None or self.pos < end:
                row = clean(next(self.iter))
                rows.append(row)
                self.pos += 1
        except StopIteration:
            self.eof = True
        return self.columns, rows

    def _fetch_from_end(self, start, end):
        if end is not None and end < 0:
            if end >= start:
                return self.columns, []
            chunk = self.df.tail(-start, compute=False, npartitions=-1)
            chunk = chunk.head(end - start, npartitions=-1)
        else:
            chunk = self.df.tail(-start, npartitions=-1)
        rows = list(map(self.clean, chunk.itertuples(index=False, name=None)))
        return self.columns, rows



import os.path

from dslibrary.sql.file_sql_wr import file_sql_write
from dslibrary.transport.to_local import DSLibraryLocal
from dslibrary.utils.dbconn import Connection
from dslibrary.utils.file_utils import is_breakout_path


def connect_to_folder_as_database(folder: str, for_write: bool=False, sql_flavor: str="mysql", **kwargs):
    """
    Each file in the folder is a table.

    :param folder:      Folder containing files.
    :param for_write:   Enable writes.
    :param sql_flavor:  Which dialect to emulate.
    :param kwargs:
    :return:        A DBI-style connection instance.
    """
    dsl = DSLibraryLocal(filesystem_root=folder)
    def file_exists(fn: str):
        return os.path.exists(os.path.join(folder, fn))
    def table_reader(table_spec: tuple):
        table_name = table_spec[-1]
        if is_breakout_path(table_name):
            return
        # allow tables to match CSV files
        # TODO what about other file formats?
        if not table_name.endswith(".csv") and file_exists(table_name+".csv"):
            table_name += ".csv"
        return dsl.load_dataframe(os.path.join(folder, table_name))
    def table_writer(table_spec: tuple, mode: str=None):
        if is_breakout_path(table_spec[-1]):
            return
        return dsl.open_resource(os.path.join(folder, table_spec[-1]), mode=mode)
    def read(sql, parameters):
        df = dsl.sql_select(sql, parameters, table_loader=table_reader)
        rows = df.itertuples(index=False, name=None)
        # TODO iteration in chunks would make more efficient use of memory - see read_more below
        return list(df.columns), list(rows), None
    def write(sql, parameters):
        file_sql_write(table_writer, sql, parameters)
    return Connection(read=read, write=write if for_write else None, read_more=None, flavor=sql_flavor)



"""
Parsing of SQL.
"""
from collections import OrderedDict

from .compiled import Node_Placeholder, Node_Literal, Node_Unary, Node_Field, Node_ExtVar, ALL_FUNCTIONS, \
    Node_Select, Node_Cast, Node_InSelect, Node_Op, Node_Op_Is, Node_List, Node_Function, Node_Case, SqlOrderBy, \
    AGGREGATE_FNS
from .exceptions import SqlException
from .tokenize import SqlTokenizer, Token
from .eval_utils import OPS_ALL, OP_PRECEDENCE, UOPS
from .select_stmt import SqlSelect


class SqlParser(object):
    """
    Parsing of SQL statements.

    Grammar/functions are similar to SQLite3.
    """
    # function names
    FN_ALL = ALL_FUNCTIONS | AGGREGATE_FNS

    def __init__(self, sql: str, allow_any_function: bool=False):
        self.input = sql
        self.tokens = self.tokenize(sql)
        self.pos = 0
        self.tables = []
        self.views = []
        self.table_map = {}
        self.view_map = {}
        self.view_filter = None
        self.allow_any_function = allow_any_function
        self.compiled_class = SqlSelect
        self.FROM_RESERVED_KEYWORDS = FROM_RESERVED_KEYWORDS[:]

    def tokenize(self, input: str):
        return SqlTokenizer(input, allowed_operators=OPS_ALL, allow_placeholders=True).run()

    def report_error(self, message):
        token = self.cur_token()
        pos = token.pos if token is not None else len(self.input)
        tkn_value = token.value if token is not None else "<end>"
        # find line and column
        lines = self.input[0:pos].split("\n")
        line = len(lines)
        col = len(lines[-1]) + 1
        b = pos - 40
        if b < 0:
            b = 0
        e = pos + 40
        if e > len(self.input):
            e = len(self.input)
        context = '"' + ("..." if b > 0 else "") + self.input[b:pos] + "<*>" + self.input[pos:e] + (
            "..." if e < len(self.input) else "") + '"'
        raise SqlParserException(message, line=line, col=col, token=tkn_value, context=context)

    def cur_token(self):
        return self.tokens[self.pos] if self.pos < len(self.tokens) else None

    def next_token(self):
        if self.pos + 1 > len(self.tokens):
            return None
        self.pos += 1
        if self.pos >= len(self.tokens):
            return None
        return self.tokens[self.pos]

    def peek_token(self, offset):
        if self.pos + offset < 0 or self.pos + offset >= len(self.tokens):
            return None
        return self.tokens[self.pos + offset]

    def _peek_test(self, offset, cmp_value):
        token = self.peek_token(offset)
        if not token:
            return False
        if cmp_value == str(token.value).lower():
            return True
        return False

    def skip_punct(self, p, error_if_missing=False):
        t = self.cur_token()
        if t and t.isPunct(p):
            self.next_token()
            return True
        if error_if_missing:
            self.report_error("Expected '" + p + "'")
        return False

    def any_keyword(self, what_its_for=None, ret_token=False):
        t = self.cur_token()
        if not t or t.type != Token.KEYWORD:
            self.report_error("Expected " + (what_its_for or "keyword"))
        self.next_token()
        return t if ret_token else t.value

    def test_keyword(self, kwd):
        """
        Check whether a particular keyword is present at the current position.
        """
        return self.skip_keyword(kwd, skip=False)

    def skip_keyword(self, kwd, skip=True):
        """
        Check whether a given keyword (or a list of keywords) is present
        at the current position.  They are skipped if they are detected and if the
        'skip' argument is not set to False.
        """
        # check for compound keyword
        if isinstance(kwd, list):
            for n in range(0, len(kwd)):
                t = self.peek_token(n)
                if not t or not t.isKeyword(kwd[n]):
                    return False
            if skip:
                self.pos += len(kwd)
            return True
        # single keyword
        t = self.cur_token()
        if t and t.isKeyword(kwd):
            if skip:
                self.next_token()
            return True
        return False

    def parse_field_name(self):
        tkn = self.any_keyword("field name", ret_token=True)
        field_name = [tkn.value]
        quotes = [bool(tkn.etc)]
        while self.skip_punct('.'):
            next_field = self.any_keyword("field component", ret_token=True)
            field_name.append(next_field.value)
            quotes.append(bool(next_field.etc))
        return tuple(field_name), tuple(quotes)

    def parse_table_name(self):
        table_name = [self.any_keyword("table name")]
        while self.skip_punct('.'):
            next_part = self.any_keyword("table component")
            table_name.append(next_part)
        return tuple(table_name)

    def parse_value(self):
        """
        <value> :=
          <number>
          <string>
          NULL
          - <value>
          '(' <expr> ')'
          <name (field)>
          <name> '(' ... ')'
        """
        t = self.cur_token()
        if not t:
            return None
        if t.isLiteral():
            self.next_token()
            if t.type == Token.PLACEHOLDER:
                return Node_Placeholder()
            return Node_Literal(value=t.value)
        elif t.value in UOPS:
            self.next_token()
            return Node_Unary(child=self.parse_value(), operator=t.value)
        elif self.skip_keyword("null"):
            return Node_Literal(value=None)
        elif self.skip_punct('('):
            return self.parse_parens()
        elif self.skip_keyword("case"):
            return self.parse_case()
        elif self.skip_keyword("cast"):
            return self.parse_cast()
        elif self.skip_punct('*'):
            # wildcard, i.e. count(*)
            return Node_Field("*")
        elif self.skip_punct('@'):
            var_name = self.any_keyword("external variable name")
            if not var_name:
                self.report_error("expected variable name after '@'")
            return Node_ExtVar(var_name)
        elif t.type == Token.KEYWORD:
            fn_call = self.parse_function_call()
            if fn_call:
                return fn_call
            # field reference
            field_name, quotes = self.parse_field_name()
            return Node_Field(field_name, quotes)
        else:
            return None

    def parse_parens(self):
        if self.test_keyword("select"):
            sel = self.parse_select()
            if not sel:
                self.report_error("expected select statement")
            expr = Node_Select(sel)
        else:
            expr = self.parse_expr_or_list()
        self.skip_punct(')', error_if_missing=True)
        return expr

    def parse_function_call(self):
        next = self.peek_token(1)
        if next and next.isPunct('('):
            t = self.cur_token()
            # function call
            function_name = t.value.lower()
            if not self.allow_any_function and function_name not in SqlParser.FN_ALL:
                self.report_error("unrecognized function: {}".format(function_name))
            self.next_token()
            self.next_token()
            if function_name == "count" and self.skip_keyword("distinct"):
                function_name = "count_distinct"
            fn = Node_Function(function_name=function_name)
            while self.cur_token():
                v = self.parse_expr()
                if not v:
                    break
                fn.children.append(v)
                if not self.skip_punct(','):
                    break
            self.skip_punct(')')
            return fn

    def parse_cast(self):
        if not self.skip_punct('('):
            self.report_error("expected '(' after 'cast'")
        expr = self.expect_expr()
        if not self.skip_keyword("as"):
            self.report_error("expected 'as'")
        to_type = self.any_keyword("type")
        if to_type == "double" and self.skip_keyword("precision"):
            to_type += " precision"
        if to_type not in {
            "integer", "int", "tinyint", "smallint", "bigint", "double", "double precision", "float", "real", "varchar",
            "date", "datetime", "timestamp", "string"
        }:
            self.report_error("expected type")
        if self.skip_punct('('):
            length = self.parse_value()
            if not isinstance(length, Node_Literal) and not isinstance(length.value, int):
                self.report_error("expected integer length in parens")
            if not self.skip_punct(')'):
                self.report_error("expected ')'")
            to_type += "(%d)" % length.value
        if not self.skip_punct(')'):
            self.report_error("expected ')'")
        return Node_Cast(expr, to_type)

    def parse_case(self):
        parts = []
        while self.skip_keyword("when"):
            parts.append(self.expect_expr())
            if not self.skip_keyword("then"):
                self.report_error("expected 'then'")
            parts.append(self.expect_expr())
        if self.skip_keyword("else"):
            parts.append(self.expect_expr())
        if not self.skip_keyword("end"):
            self.report_error("expected 'end'")
        return Node_Case(parts)

    def parse_op(self):
        t = self.cur_token()
        if not t:
            return None
        if t.type not in [Token.PUNCT, Token.KEYWORD]:
            return None
        op = t.value.lower()
        if op not in OPS_ALL:
            return None
        self.next_token()
        return op

    def expect_expr(self):
        """
        Parse an expression and fail if none is found.
        """
        expr = self.parse_expr()
        if not expr:
            self.report_error("expected expression")
        return expr

    def parse_expr_or_list(self):
        v = self.parse_expr()
        if not v:
            return None
        if self.cur_token() and self.cur_token().value == ",":
            the_list = Node_List()
            the_list.children.append(v)
            while self.skip_punct(","):
                v = self.parse_expr()
                if not v:
                    self.report_error("expecting value after ','")
                the_list.children.append(v)
            return the_list
        return v

    def parse_expr(self):
        v_stack = []
        op_stack = []
        while True:
            v = self.parse_value()
            if not v:
                if v_stack:
                    self.report_error("expected value")
                else:
                    return v
            if self.skip_keyword("between"):
                v = self.parse_between(v)
            v_stack.append(v)
            op = self.parse_op()
            if not op:
                break
            while op_stack and OP_PRECEDENCE[op_stack[-1]] > OP_PRECEDENCE[op]:
                v_stack[-2] = self._expr_build_op(v_stack[-2], op_stack[-1], v_stack[-1])
                v_stack.pop(-1)
                op_stack.pop(-1)
            op_stack.append(op)
        while op_stack:
            v_stack[-2] = self._expr_build_op(v_stack[-2], op_stack[-1], v_stack[-1])
            v_stack.pop(-1)
            op_stack.pop(-1)
        return v_stack[0] if v_stack else None

    def parse_between(self, v_cmp):
        v_l = self.parse_value()
        if not v_l:
            self.report_error("expected value after 'between'")
        if not self.skip_keyword("and"):
            self.report_error("expected 'and' between between values")
        v_r = self.parse_value()
        if not v_r:
            self.report_error("expected value after 'between'")
        return self._expr_build_op(self._expr_build_op(v_cmp, ">=", v_l), "and", self._expr_build_op(v_cmp, "<=", v_r))

    @staticmethod
    def _expr_build_op(v_l, op, v_r):
        if op in ("in", "not in"):
            if isinstance(v_r, Node_Select):
                o = Node_InSelect(v_l, v_r.select_spec)
                if op == "not in":
                    o = Node_Unary(child=o, operator="not")
            else:
                o = Node_Op(op, [v_l, v_r])
            return o
        if op in ("is", "is not"):
            o = Node_Op_Is(op)
        else:
            o = Node_Op(op)
        o.children.append(v_l)
        o.children.append(v_r)
        return o

    def parse_from_expr(self):
        """
        A specification for a source table, with aliases and joins.

        tableExpr :=
          table (alias) (( jointype table2 (alias2) 'on' condition ))

        jointype := (( inner | outer | right | left )) join
        """
        out = SqlSelect.FromExpr()
        out.table, out.alias, out.subquery = self._table_and_alias()
        tbl = out
        if self.peek_token(0) and self.peek_token(0).isPunct(','):
            # old style joins: table1, table2, ...
            while self.skip_punct(','):
                tbl.join = SqlSelect.Join()
                tbl = tbl.join
                tbl.table, tbl.alias, tbl.subquery = self._table_and_alias()
                tbl.joinType = {"cross"}
            return out
        # new style joins
        while self.cur_token():
            join = SqlSelect.Join()
            join.joinType = set()
            while self.cur_token() and self.cur_token().type == Token.KEYWORD and self.cur_token().value.lower() in FROM_JOIN_KEYWORDS:
                join.joinType.add(self.any_keyword("join type").lower())
            if not join.joinType:
                break
            join.table, join.alias, join.subquery = self._table_and_alias()
            if self.skip_keyword("on"):
                join.joinExpr = self.expect_expr()
            tbl.join = join
            tbl = join
        return out

    def _table_and_alias(self):
        subquery = None
        table = None
        if self._peek_test(0, "(") and self._peek_test(1, "select"):
            self.skip_punct("(")
            subquery = self.parse_select()
            if not subquery:
                self.report_error("expected select statement")
            self.skip_punct(")", error_if_missing=True)
        else:
            table = self.parse_table_name()
        alias = None
        if self.skip_keyword("as"):
            alias = self.any_keyword("alias")
        elif self.cur_token() and self.cur_token().type == Token.KEYWORD and self.cur_token().value.lower() not in self.FROM_RESERVED_KEYWORDS:
            alias = self.any_keyword("alias")
        return table, alias, subquery

    def parse_keyword_list(self):
        while True:
            yield self.any_keyword("field name")
            if not self.skip_punct(','):
                break

    def _parse_field_for_select(self, add_to_list):
        # detect '*' wildcard
        if self.skip_punct('*'):
            wildcard = "*"
            if self.skip_punct('?'):
                # TODO I think this was some kind of SQL extension but I can't find support for it elsewhere
                wildcard += "?"
            elif self.skip_punct('['):
                # custom SQL extension: all fields EXCEPT those listed here
                wildcard += "["
                wildcard += ",".join(self.parse_keyword_list())
                wildcard += "]"
                self.skip_punct(']', True)
            add_to_list.append((Node_Field((None, wildcard,)), None))
            return True
        # detect (table).* wildcard
        if self.peek_token(0).type == Token.KEYWORD and self.peek_token(2) and self.peek_token(
                1).value == "." and self.peek_token(2).value == '*':
            table_name = self.cur_token().value
            self.next_token()
            self.next_token()
            self.next_token()
            add_to_list.append((Node_Field((table_name, "*")), None))
            return True
        # otherwise, a field is an expression plus an optional alias
        calc = self.parse_expr()
        alias = None
        if not calc:
            return False
        if self.skip_keyword("as"):
            alias = self.any_keyword("field alias")
        elif self.cur_token() and self.cur_token().type == Token.KEYWORD and self.cur_token().value.lower() not in {
                'from', 'where', 'order', 'limit'
        }:
            alias = self.cur_token().value
            self.next_token()
        add_to_list.append((calc, alias))
        return True

    def parse_fields(self):
        """ Parses the part of a SELECT that defines the fields to extract """
        fields = []
        while self.cur_token():
            if not self._parse_field_for_select(fields):
                self.report_error("Expected field spec")
            if not self.skip_punct(','):
                break
        return fields

    def parse_group_by(self):
        if self.skip_keyword("group"):
            self.skip_keyword("by")
            group_by = SqlOrderBy()
            while self.cur_token():
                expr = self.expect_expr()
                group_by.addLevel(expr)
                if not self.skip_punct(','):
                    break
            return group_by

    def parse_order_by(self):
        if self.skip_keyword("order"):
            self.skip_keyword("by")
            order_by = SqlOrderBy()
            while self.cur_token():
                expr = self.expect_expr()
                asc = True
                if self.skip_keyword("ascending") or self.skip_keyword("asc"):
                    # valid
                    pass
                elif self.skip_keyword("descending") or self.skip_keyword("desc"):
                    asc = False
                order_by.addLevel(expr, asc)
                if not self.skip_punct(','):
                    break
            return order_by

    def parse_list_of_fields(self):
        fields = []
        if self.skip_punct('('):
            while True:
                if self.skip_punct(')'):
                    break
                # TODO quoting style is preserved in all cases but this one here
                field_names, quotes = self.parse_field_name()
                fields.append(field_names)
                if self.skip_punct(')'):
                    break
                self.skip_punct(',')
        return fields

    def _keyword_then(self, keyword: str, fn: callable, skip: bool=True):
        """
        A very useful shorthand.
        :param keyword:     Keyword to test for.
        :param fn:          Code to execute if it is found.
        :return:            Return value from fn(), or None.
        """
        if self.skip_keyword(keyword, skip=skip):
            expr = fn()
            return expr

    def _terms(self):
        return [
            ("from_expr", lambda: self._keyword_then("from", self.parse_from_expr)),
            ("where", lambda: self._keyword_then("where", self.expect_expr)),
            ("group_by", lambda: self._keyword_then("group", self.parse_group_by, skip=False)),
            ("having", lambda: self._keyword_then("having", self.expect_expr)),
            ("order_by", lambda: self._keyword_then("order", self.parse_order_by, skip=False)),
            ("start", lambda: self._keyword_then("offset", self.expect_expr)),
            ("limit", lambda: self._keyword_then("limit", self.expect_expr))
        ]

    def parse_select(self):
        """
        A 'select' statement.
        """
        if not self.skip_keyword("select"):
            return None
        out = self.compiled_class()
        if self.skip_keyword("distinct"):
            out.distinct = self.parse_list_of_fields()
        # parse field list
        out.fields = self.parse_fields()
        # process each part of the statement
        for attr, capture in self._terms():
            setattr(out, attr, capture())
        # selects can be combined with 'union'
        while self.skip_keyword("union"):
            out.union = []
            sel = self.parse_select()
            if not sel:
                self.report_error("expected SELECT")
            out.union.append(sel)
        return out

    def parse_select_only(self):
        """
        Parse a select statement and raise an exception if one is not found.  Also verifies that there is
        no garbage after the statement.
        """
        sel = self.parse_select()
        if not sel:
            self.report_error("Expected SELECT statement")
        self.skip_punct(";")
        if self.cur_token():
            self.report_error("Syntax error after statement")
        return sel

    def parse_read_statement(self):
        """
        Extension point for parsers that support read statements other than SELECT, like SHOW or DESCRIBE.
        """
        return self.parse_select_only()


def _get_select(sql_or_select, alt_parser=None):
    """
    Parses a select statement, or recognizes that the parameter is an already-parsed select statement.
    """
    if isinstance(sql_or_select, SqlSelect):
        return sql_or_select
    if sql_or_select not in compiler_cache_1:
        p = (alt_parser or SqlParser)(sql_or_select)
        parsed = p.parse_read_statement()
        compiler_cache_1[sql_or_select] = parsed
    return compiler_cache_1.get(sql_or_select)


def _lookup_tables_for_field_spec(select, field_spec, alias_to_table):
    """
    Find the tables associated with a wildcard.
    """
    if len(field_spec) > 1 and field_spec[-2]:
        # look up the table qualifying the wildcard
        table_ref = field_spec[:-1]
        tables = [table_ref]
        if not alias_to_table:
            alias_to_table = select.getAliasToTableMapping()
        if tables[0][-1] in alias_to_table:
            tables[0] = alias_to_table[tables[0][-1]]
    else:
        # search all tables in statement
        tables = find_referenced_tables(select, include_union=False, include_subselect=False, include_subquery=True)
    return tables or []


def expand_field_spec(select, field_spec, get_fields_for_table, alias_to_table=None):
    """
    Given the field specification of Node_Field, i.e. a tuple representing the parts of a
    reference to a field (i.e. a.b --> ("a","b")), look up that field and expand it into a consistent
    field specification, with fully normalized database, table and field components.  Aliases for field
    and table names are turned into proper field names, and wildcards are expanded.

    :param select:            A parsed SELECT statement.
    :param field_spec:         A tuple with at least a field name.
    :param get_fields_for_table: A method that will return a list of column names, given a table specification tuple.
    :param alias_to_table:      The alias-to-table mapping returned from SqlSelect (optional, to improve performance).
    :return:        A list of field specification 3-tuples.
    """
    if not field_spec[-1].startswith("*"):
        # just look up the field
        expanded = select.findFieldRef(field_spec, get_fields_for_table)
        return [expanded or field_spec]
    tables = _lookup_tables_for_field_spec(select, field_spec, alias_to_table=alias_to_table)
    out = []
    for table in tables:
        # 'table' can be the alias for a subquery
        sub = select.findSubqueryByAlias(table[-1])
        cols = []
        if sub:
            # descend to the subquery and enumerate its fields
            cols = expand_field_spec(sub, ("*",), get_fields_for_table)
            cols = [c[-1] for c in cols if c[-1]]
        elif get_fields_for_table:
            # or it can be the name of a real table, which we need to enumerate
            cols = get_fields_for_table(table) or []
        out += [table + (col,) for col in cols]
        # represent the wildcard/passthrough with a "" for the field name
        out.append(table + ("",))
    # custom language extension: the brackets let you exclude particular field names from the wildcard
    if field_spec[-1].startswith("*["):
        exclude = field_spec[-1][2:-1].split(',')
        out = filter(lambda x: x[-1] not in exclude, out)
    return out


def find_one_to_one_field_mappings(sql, get_fields_for_table=None):
    """
    For the given SQL, returns a mapping from output field name to a tuple describing
    the associated input field.  The tuples contain three components: database, table and field.

    Instead of a tuple, None will be mapped when there is no single field associated with the output.

    Returns None if the SQL did not contain a SELECT statement.

    There are a number of cases where getFieldsForTable is required in order for the results to be complete.
    These include when there are any wildcards and when there are multiple tables and fields without table
    qualifiers.
    """
    sel = _get_select(sql)
    out = {}
    alias_to_table = sel.getAliasToTableMapping()
    for field_spec in sel.fields:
        if not isinstance(field_spec[0], Node_Field):
            output_field_name = field_spec[1] or str(field_spec[0])
            out[output_field_name] = (None, None)
            continue
        expanded = expand_field_spec(sel, field_spec[0].fieldName, get_fields_for_table, alias_to_table=alias_to_table)
        if len(expanded) == 1:
            output_field_name = field_spec[1] or expanded[0][-1]
            out[output_field_name] = expanded[0]
            continue
        out.update({f[-1]: f for f in expanded if f[-1] is not None})
    return out


def find_output_columns(sql, get_fields_for_table=None):
    """
    Given a SELECT statement, return a list of all columns that it produces.
    """
    full_mapping = find_one_to_one_field_mappings(sql, get_fields_for_table)
    return full_mapping.keys()


class TableRefFinder(object):
    def __init__(self, include_union=True, include_subselect=True, include_subquery=True, alt_parser=None):
        self.include_union = include_union
        self.include_subselect = include_subselect
        self.include_subquery = include_subquery
        self.alt_parser = alt_parser
        self.out = OrderedDict()

    def refs_in_node(self, node):
        if not self.include_subselect:
            return
        if isinstance(node, Node_Select) or isinstance(node, Node_InSelect):
            self.refs_in_select(node.select_spec)
        elif hasattr(node, "children"):
            for sub in node.children:
                self.refs_in_node(sub)

    def refs_in_select(self, sel):
        # extensions provide a method to enumerate referenced tables
        if not isinstance(sel, SqlSelect):
            for tbl in sel.find_all_tables():
                self.out[tbl] = 1
            return
        tbl = sel.from_expr
        while tbl:
            if tbl.table:
                self.out[tbl.table] = 1
            elif tbl.subquery and self.include_subquery:
                self.refs_in_select(tbl.subquery)
            tbl = tbl.join
        if self.include_subselect:
            # NOTE: includeSubselect currently spans unions and subqueries whether you ask for it or not
            for expr in sel.allExpressions():
                self.refs_in_node(expr)
        if sel.union and self.include_union:
            for u in sel.union:
                self.refs_in_select(u)

    def find(self, sql):
        sel = _get_select(sql, alt_parser=self.alt_parser)
        self.refs_in_select(sel)
        return list(self.out.keys())


def find_referenced_tables(sql, include_union=True, include_subselect=True, include_subquery=True, alt_parser=None):
    """
    Given a SELECT statement, return a list of all tables that it references.

    :returns:  A list of tuples.  Each tuple contains the name of a database and the name of a table.  The
               databases will often be None, and the tables should never be None.
    """
    return TableRefFinder(include_union, include_subselect, include_subquery, alt_parser=alt_parser).find(sql)


class _FieldRefFinder(object):
    def __init__(self, get_fields_for_table=None):
        self.get_fields_for_table = get_fields_for_table
        self.out = set()

    def refs_in_node(self, select, node, alias_to_table):
        if isinstance(node, Node_Field):
            ref = select.findFieldRef(node.fieldName, self.get_fields_for_table)
            if not ref:
                ''' this is either an internal error in the SQL, or we don't have getFieldsForTable filled in,
                    either way we can't figure out which field is being referenced
                ref = node.fieldName
                if len(ref) == 1:
                    ref = (None,) + ref
                '''
                return
            expanded = expand_field_spec(select, ref, self.get_fields_for_table, alias_to_table=alias_to_table)
            for f in expanded:
                self.out.add(f)
        elif isinstance(node, Node_InSelect):
            self.refs_in_select(node.select_spec)
        if hasattr(node, "children"):
            for sub in node.children:
                self.refs_in_node(select, sub, alias_to_table)

    def refs_in_select(self, select):
        alias_to_table = select.getAliasToTableMapping()
        for expr in select.allExpressions(includeUnions=False, includeSubqueries=False):
            self.refs_in_node(select, expr, alias_to_table)
        tbl = select.from_expr
        while tbl:
            if tbl.subquery:
                self.refs_in_select(tbl.subquery)
            tbl = tbl.join
        if select.union:
            for u in select.union:
                self.refs_in_select(u)

    def find(self, sql: str):
        sel = _get_select(sql)
        self.refs_in_select(sel)
        # eliminate references to field aliases
        field_aliases = {fieldInfo[1] for fieldInfo in sel.fields if _field_spec_is_alias_only(fieldInfo)}
        if field_aliases:
            self.out = set(filter(lambda r: r[-1] not in field_aliases, self.out))
        return self.out


def find_referenced_fields(sql, get_fields_for_table=None):
    """
    Determine all input fields that are referenced in a given statement.  Returns 3-tuples for each field,
    with database, table and field names.
    """
    return _FieldRefFinder(get_fields_for_table=get_fields_for_table).find(sql)


def _expr_contains_field(expr, field_name):
    if isinstance(expr, Node_Field):
        if expr.fieldName[-1] == field_name:
            return True
    elif hasattr(expr, "children"):
        for child in expr.children:
            if _expr_contains_field(child, field_name):
                return True


def _field_spec_is_alias_only(field_spec):
    if not field_spec[1]:
        return False
    if isinstance(field_spec[0], Node_Field):
        if field_spec[0].fieldName[-1] == field_spec[1]:
            return False
        return True
    else:
        # If there is a field reference that matches the alias, the alias is only replacing the original field name
        #  and the original field name is still a valid field name.
        if _expr_contains_field(field_spec[0], field_spec[1]):
            return False
    return True



def find_excluded_wildcard_outputs(sql):
    """
    Find any fields that are specifically excluded from wildcard field selection.
    That is, for SQL like "select *[f1,f2] from t", it would return [f1,f2].
    This list would always be empty for normal SQL.
    :param sql: SQL to analyze (a SELECT statement).
    :return:    List of field names which will be excluded.
    """
    sel = _get_select(sql)
    out = set()
    for f in sel.fields:
        if not isinstance(f[0], Node_Field):
            continue
        if f[0].isWildcard():
            for f in f[0].getWildcardExceptions() or []:
                out.add(f)
        else:
            fn = f[0].fieldName
            if fn in out:
                out.remove(fn)
    return out


class SqlParserException(SqlException):
    def __init__(self, message, context, line, col, token):
        super(SqlParserException, self).__init__(message)
        self.context = context
        self.line = line
        self.col = col
        self.token = token

    def __str__(self):
        return u"{message}, line:col={line}:{col}, token={token}, context={context}".format(**self.__dict__)


# to speed up compilation
compiler_cache_1 = {}
# TODO get this working, currently it creates a recursive import problem
# compiler_cache_1 = LruCache(maxSize=1000)

FROM_JOIN_KEYWORDS = ["inner", "outer", "right", "left", "join", "full", "natural", "cross"]
FROM_RESERVED_KEYWORDS = FROM_JOIN_KEYWORDS + ["where", "union", "order", "offset", "limit", "group", "having", "on"]



class SqlException(Exception):
    def __init__(self, message=None, **kwargs):
        Exception.__init__(self, message)
        self.message = message
        self.details = kwargs


class SqlException_SqlProblem(SqlException):
    """
    Problems with caller-supplied SQL.
    """
    def __init__(self, message=None, **kwargs):
        SqlException.__init__(self, message, **kwargs)


class SqlException_DataSourceNotFound(SqlException):
    def __init__(self, message=None):
        SqlException.__init__(self, message)


class SqlException_NotImplemented(SqlException):
    def __init__(self, message=None):
        SqlException.__init__(self, message)



"""
Interpret sql-to-pandas errors.
"""
import re
import sys
import traceback

from dslibrary import DSLibraryDataFormatException

PTN_TITLE = re.compile(r'^(.*?)\s+Traceback \(most recent call last\)$')
PTN_LEVEL = re.compile(r'^([^\n]+) in ([^\n]+)\n(.*)$', re.DOTALL)
PTN_DETAIL_W_TRACE = re.compile(r'^([^\n]+)\n\nOriginal error is below:\n----.*$', re.DOTALL)
PTN_ANOTHER = re.compile(r'\nDuring handling of the above exception, another exception occurred:\n')
PTN_REF_LINE = re.compile(r'^-+>\s*(\d+)\s+(.*)$')


def parse_ipykernel_traceback(lines: list):
    """
    Tracebacks are returned in a 'sort of visual' console format.  Here we attempt to turn it into a reasonably
    machine-interpretable format.
    """
    if not lines:
        return
    out = {"error": "", "levels": [], "detail": ""}
    current = out
    for line in lines:
        if line.startswith("------------"):
            continue
        m = PTN_TITLE.match(line)
        if m is not None:
            current["error"] = m.group(1)
            continue
        if PTN_ANOTHER.match(line) is not None:
            next = current["cause"] = {"error": "", "levels": [], "detail": ""}
            current = next
            continue
        m = PTN_DETAIL_W_TRACE.match(line)
        if m is not None:
            current["detail"] += m.group(1).strip() + "\n"
            continue
        level = _parse_tb_level(line)
        if level:
            current["levels"].append(level)
            continue
        current["detail"] += line.replace("\\n", "\n").strip() + "\n"
    return out


def _parse_tb_level(line: str):
    m = PTN_LEVEL.match(line)
    if m is not None:
        ref_line = list(filter(lambda l: l.startswith("-"), m.group(3).split("\n")))
        mref = PTN_REF_LINE.match(ref_line[0]) if ref_line else None
        return {
            "file": m.group(1),
            "method": m.group(2),
            "line": int(mref.group(1)) if mref is not None else None,
            "code": mref.group(2) if mref is not None else None
        }


def ipykernel_exc_to_json(exc_info=None) -> dict:
    """
    Convert an in-scope exception into a format that corresponds to what ipykernel returns, after its traceback is parsed by
    parse_ipykernel_traceback().
    """
    err_type, err_value, err_traceback = (exc_info or sys.exc_info())
    tb = traceback.extract_tb(err_traceback)
    return {
        "ename": err_type.__name__,
        "evalue": str(err_value),
        "traceback": {
            "error": err_type.__name__,
            "detail": str(err_value),
            "levels": [
                {"file": level.filename, "method": level.name, "line": level.lineno, "code": level.line or ""}
                for level in tb
            ]
        }
    }


def interpret_s2p_errors(err: dict):
    """
    Detect common SQL errors and add their description into a traceback.
    """
    _detect_datatype_conflict(err)


def _detect_datatype_conflict(err: dict):
    if err["ename"] not in {"ValueError", "TypeError"}:
        return
    if err["evalue"].startswith("Metadata inference failed in"):
        # dask
        pass
    elif "not supported between instances of " in err["evalue"]:
        # pandas
        pass
    else:
        return
    err["interpretation"] = \
        "The data types in the requested operation are not compatible.\n" \
        "You may need to clean one or more columns in order that they consistently contain the intended data type.\n"\
        "Reported error was: %s" % err["traceback"]["detail"].strip()


def translate_pandas_dask_exceptions(exc):
    """
    Recognize some exceptions and provide more helpful information/codes.
    """
    if isinstance(exc, UnicodeDecodeError):
        return DSLibraryDataFormatException(f"Encoding problem in file, reason={exc.reason}, offset={exc.start}, encoding={exc.encoding}")
    s_exc = str(exc)
    if "Error tokenizing data." in s_exc or ("Expected " in s_exc and "fields in line " in s_exc):
        return DSLibraryDataFormatException(f"CSV problem: {s_exc}")
    if isinstance(exc, ValueError) and ("Trailing data" in s_exc or "Expected object" in s_exc or "No ':'" in s_exc):
        return DSLibraryDataFormatException(f"JSON problem: {s_exc}")
    if isinstance(exc, ValueError):
        return DSLibraryDataFormatException(f"Data format problem: {s_exc}")
    return exc



"""
Simple SQL modifications, applied to files.
"""
import io
import pandas

import dslibrary
from dslibrary.sql.exceptions import SqlException
from dslibrary.sql.modify_sql import embed_parameters
from dslibrary.sql.parser import SqlParser
from dslibrary.sql.sql_to_pandas import SqlToPandas
from dslibrary.sql.statements import parse_insert, parse_delete


def file_sql_write(open_table_stream: callable, sql: str, parameters: list=None, format_options=None):
    """
    Run SQL against files.
    :param open_table_stream:   Method to open files for read/write/etc..
    :param sql:                 SQL to execute.
    :param parameters:          Parameters to fill into SQL.
    :param format_options:      Format details.
    """
    sql = embed_parameters(sql, parameters)
    parser = SqlParser(sql)
    if parser.skip_keyword("CREATE"):
        if parser.skip_keyword("TABLE"):
            return fsql_create_table(parser, open_table_stream)
        if parser.skip_keyword("INDEX"):
            """ ignoring CREATE TABLE """
            return
    if parser.skip_keyword("INSERT"):
        return fsql_insert(sql, open_table_stream, format_options=format_options)
    if parser.skip_keyword("DELETE"):
        return fsql_delete(sql, open_table_stream, format_options=format_options)
    if parser.skip_keyword("UPDATE"):
        """ not supported yet: UPDATE """
        pass
    elif parser.skip_keyword("DROP"):
        if parser.skip_keyword("TABLE"):
            """ not supported yet: DROP TABLE """
            pass
        if parser.skip_keyword("INDEX"):
            """ ignoring DROP INDEX """
            return
        pass
    elif parser.skip_keyword("ALTER"):
        """ not supported yet: ALTER TABLE """
        pass
    parser.report_error(f"unsupported SQL: {sql}")


def fsql_create_table(parser, opener):
    """
    CREATE TABLE
    """
    if_not_exist = False
    if parser.skip_keyword("IF"):
        if not parser.skip_keyword("NOT"):
            parser.report_error("invalid create table syntax")
        if not parser.skip_keyword("EXISTS"):
            parser.report_error("invalid create table syntax")
        if_not_exist = True
    table_name = parser.any_keyword("table name")
    parser.skip_punct("(", error_if_missing=True)
    fields = []
    while True:
        field_name = parser.any_keyword("field name")
        fields.append(field_name)
        # (we ignore the data type)
        dt = parser.any_keyword("data type")
        if dt == "double" and parser.skip_keyword("precision"):
            # allow 'double precision'
            dt += " precision"
        if parser.skip_punct("("):
            parser.parse_value()
            if not parser.skip_punct(')'):
                parser.report_error("Expected ')' after data type qualifier")
        if parser.skip_keyword("NULL"):
            # we ignore nullability
            pass
        elif parser.skip_keyword("NOT") and not parser.skip_keyword("NULL"):
            parser.report_error("expected 'nullable' after 'not'")
        if not parser.skip_punct(','):
            break
    if not parser.skip_punct(')'):
        parser.report_error("expected ')' after field list")
    # flag to skip duplicate creation
    if if_not_exist:
        try:
            with opener((table_name,), 'rb') as found:
                # exists
                pass
            if found:
                return
        except Exception:
            # failure means file doesn't exist
            pass
    # creation: store a header
    with opener((table_name,), 'w') as f_w:
        content = "\t".join(fields) + "\n"
        f_w.write(content)


def fsql_insert(sql, opener, format_options=None):
    """
    INSERT INTO
    """
    table_name, fields, rows = parse_insert(sql)
    # verify table exists, read headers
    f_r = opener(table_name, 'rb')
    if not f_r:
        table = pandas.DataFrame(data=rows, columns=fields)
    else:
        with f_r:
            table = dslibrary.load_dataframe(f_r, uri=table_name[-1], format_options=format_options)
        table = pandas.concat([table, pandas.DataFrame(rows, columns=fields)])
    buf = io.StringIO()
    table.to_csv(buf, sep="\t", index=False)
    with opener(table_name, 'w') as f_w:
        f_w.write(buf.getvalue())


def fsql_delete(sql, opener, format_options=None):
    """
    DELETE rows
    """
    table_name, where = parse_delete(sql)
    # verify table exists, read headers
    f_r = opener(table_name, 'rb')
    if not f_r:
        raise SqlException(message=f"table not found: {table_name}")
    with f_r:
        table = dslibrary.load_dataframe(f_r, uri=table_name[-1], format_options=format_options)
    # delete rows
    if where:
        ctx = {"tables": {"table": table}}
        to_run = compile(SqlToPandas(f"select * from table where not({where})", local_tables="tables", pandas_only=True).code, "<code>", "exec")
        eval(to_run, ctx, ctx)
        table = ctx["_"]
    else:
        table = table[0:0]
    # regenerate table content
    buf = io.StringIO()
    table.to_csv(buf, sep="\t", index=False)
    with opener(table_name, 'w') as f_w:
        f_w.write(buf.getvalue())
